import numpy as np
import sys
import torch
import math
import random

"""
Below are some default wind objects that might be useful inputs to the system.  
"""

class NoWind(object):
    """
    This wind profile is the trivial case of no wind. It will output 
    zero wind speed on all axes for all time. 
    Alternatively, you can use ConstantWind with wx=wy=wz=0. 
    """

    def __init__(self):
        """
        Inputs: 
            Nothing
        """
        self.wind = np.array([0, 0, 0])

    def update(self, t, position):
        """
        Given the present time and position of the multirotor, return the
        current wind speed on all three axes. 
        
        The wind should be expressed in the world coordinates.
        """
        return self.wind

class BatchedNoWind(object):
    '''
    Batched version of NoWind that prepends a batch dimension to the output.
    '''
    def __init__(self, num_drones):
        self.wind = torch.zeros(num_drones, 3)

    def update(self, t, position):
        return self.wind

class ConstantWind(object):
    """
    This wind profile is constant both spatially and temporally. 
    Wind speed is specified on each axis. 
    """

    def __init__(self, wx, wy, wz):
        """
        """
        self.wind = np.array([wx, wy, wz])

        
    def update(self, t, position):
        """
        Given the present time and position of the multirotor, return the
        current wind speed on all three axes. 
        
        The wind should be expressed in the world coordinates. 
        """

        return self.wind

class BatchedConstantWind(object):
    """
    Batched version of ConstantWind that prepends a batch dimension to the output.
    """
    def __init__(self, num_drones, wx, wy, wz):
        self.wind = torch.tensor([wx, wy, wz]).repeat(num_drones, 1)

    def update(self, t, position):
        return self.wind


class SinusoidWind(object):
    """
    Wind will vary subject to a sine function with appropriate amplitude, frequency, and phase offset.
    """

    def __init__(self, amplitudes=np.array([1,1,1]), frequencies=np.array([1,1,1]), phase=np.array([0,0,0])):
        """
        Inputs: 
            amplitudes := array of amplitudes on each axis
            frequencies := array of frequencies for the wind pattern on each axis
            phase := relative phase offset on each axis in seconds
        """
        self.Ax, self.Ay, self.Az = amplitudes[0], amplitudes[1], amplitudes[2]
        self.fx, self.fy, self.fz = frequencies[0], frequencies[1], frequencies[2]
        self.px, self.py, self.pz = phase[0], phase[1], phase[2]

    def update(self, t, position):
        """
        Given the present time and position of the multirotor, return the
        current wind speed on all three axes. 
        
        The wind should be expressed in the world coordinates. 
        """

        wind = np.array([self.Ax*np.sin(2*np.pi*self.fx*(t+self.px)),
                         self.Ay*np.sin(2*np.pi*self.fy*(t+self.py)),
                         self.Az*np.sin(2*np.pi*self.fz*(t+self.pz))])

        return wind

class BatchedSinusoidWind(object):
    """
    Batched version of SinusoidWind that handles a batch of UAV, each with
    their own amplitude, frequency, and phase on each axis.
    """

    def __init__(self, amplitudes, frequencies, phases):
        """
        Inputs:
            amplitudes := tensor of shape (num_drones, 3) for amplitude per UAV per axis
            frequencies := tensor of shape (num_drones, 3) for frequency per UAV per axis
            phases := tensor of shape (num_drones, 3) for phase per UAV per axis (in seconds)
        """
        self.A = amplitudes
        self.f = frequencies
        self.p = phases

    def update(self, t, position):
        """
        Inputs:
            t := scalar float time
            position := tensor of shape (num_drones, 3), not used in this wind model
        Returns:
            wind := tensor of shape (num_drones, 3)
        """

        t_tensor = torch.full_like(self.p, t)
        wind = self.A * torch.sin(2*math.pi*self.f*(t_tensor + self.p))
        return wind

class LadderWind(object):
    """
    The wind will step up and down between a minimum and maximum speed for a specified duration. 
    Visualized below...

                     | | <- duration
     max ->          ---         ---
                  ---         ---
               ---         ---
     min -> ---         ---
            --------------------------> t
    
    ** Normally the wind will start at min and increase sequentially, but if the random flag is set true, 
       the wind will step to a random sublevel after each duration is up. 

    """

    def __init__(self, min=np.array([-1,-1,-1]), max=np.array([1,1,1]), duration=np.array([1,1,1]), Nstep=np.array([5,5,5]), random_flag=False):
        """
        Inputs:
            min := array of minimum wind speeds across each axis
            max := array of maximum wind speeds across each axis
            duration := array of durations for each step
            Nstep := array for the integer number of discretized steps between min and max across each axis
        """

        # Check the inputs for consistency, quit and raise a flag if the inputs aren't physically realizable
        if np.any(Nstep <= 0):
            print("LadderWind Error: The number of steps must be greater than or equal to 1")
            sys.exit(1)

        if np.any(max - min < 0):
            print("LadderWind Error: The max value must be greater than the min value.")
            sys.exit(1)

        self.random_flag = random_flag
        self.durx, self.dury, self.durz = duration[0], duration[1], duration[2]
        self.nx, self.ny, self.nz = Nstep[0], Nstep[1], Nstep[2]

        # Compute arrays of intermediate wind speeds for each axis
        self.wx_arr = np.linspace(min[0], max[0], self.nx)
        self.wy_arr = np.linspace(min[1], max[1], self.ny)
        self.wz_arr = np.linspace(min[2], max[2], self.nz)

        # Initialize the amplitude id.. these numbers are used to index the arrays above to get the appropriate wind speed on each axis
        if self.random_flag:
            self.xid = np.random.choice(self.nx)
            self.yid = np.random.choice(self.ny)
            self.zid = np.random.choice(self.nz)
        else:
            self.xid, self.yid, self.zid = 0, 0, 0

        # Initialize the timers... since we don't yet know the starting time, we'll set them in the first call
        self.timerx = None
        self.timery = None
        self.timerz = None

        # Initialize the winds
        self.wx, self.wy, self.wz = self.wx_arr[self.xid], self.wy_arr[self.yid], self.wz_arr[self.zid]

    def update(self, t, position):
        """
        Given the present time and position of the multirotor, return the
        current wind speed on all three axes. 
        
        The wind should be expressed in the world coordinates. 
        """
        if self.timerx is None:
            self.timerx, self.timery, self.timerz = t, t, t
        
        if (t - self.timerx) >= self.durx:
            if self.random_flag:
                self.xid = np.random.choice(self.nx)
            else:
                self.xid = (self.xid + 1) % self.nx
            self.wx = self.wx_arr[self.xid]
            self.timerx = t

        if (t - self.timery) >= self.dury:
            if self.random_flag:
                self.yid = np.random.choice(self.ny)
            else:
                self.yid = (self.yid + 1) % self.ny
            self.wy = self.wy_arr[self.yid]
            self.timery = t

        if (t - self.timerz) >= self.durz:
            if self.random_flag:
                self.zid = np.random.choice(self.nz)
            else:
                self.zid = (self.zid + 1) % self.nz
            self.wz = self.wz_arr[self.zid]
            self.timerz = t

        return np.array([self.wx, self.wy, self.wz])

class BatchedLadderWind(object):
    """
    Batched version of LadderWind. Each drone has its own wind ladder across all axes.
    """

    def __init__(self, min_vals, max_vals, durations, nsteps, random_flag=False):
        """
        Inputs:
            min_vals := tensor of shape (num_drones, 3)
            max_vals := tensor of shape (num_drones, 3)
            durations := tensor of shape (num_drones, 3)
            nsteps := tensor of shape (num_drones, 3), integer number of steps between min and max
        """
        self.random_flag = random_flag
        self.num_drones = min_vals.shape[0]  # Get number of UAVs

        if torch.any(nsteps <= 0):
            raise ValueError("LadderWind Error: The number of steps must be greater than or equal to 1")

        if torch.any(max_vals - min_vals < 0):
            raise ValueError("LadderWind Error: The max value must be greater than the min value.")

        self.durations = durations  # shape (num_drones, 3)
        self.nsteps = nsteps        # shape (num_drones, 3)

        # Precompute the ladder values for each axis per drone
        self.wx_ladders = [torch.linspace(min_vals[i, 0], max_vals[i, 0], int(nsteps[i, 0])) for i in range(self.num_drones)]
        self.wy_ladders = [torch.linspace(min_vals[i, 1], max_vals[i, 1], int(nsteps[i, 1])) for i in range(self.num_drones)]
        self.wz_ladders = [torch.linspace(min_vals[i, 2], max_vals[i, 2], int(nsteps[i, 2])) for i in range(self.num_drones)]

        # Initialize step indices
        if random_flag:
            self.xids = torch.tensor([random.randint(0, len(ladder) - 1) for ladder in self.wx_ladders])
            self.yids = torch.tensor([random.randint(0, len(ladder) - 1) for ladder in self.wy_ladders])
            self.zids = torch.tensor([random.randint(0, len(ladder) - 1) for ladder in self.wz_ladders])
        else:
            self.xids = torch.zeros(self.num_drones, dtype=torch.long)
            self.yids = torch.zeros(self.num_drones, dtype=torch.long)
            self.zids = torch.zeros(self.num_drones, dtype=torch.long)

        # Timers (initialized on first update)
        self.timers_initialized = False
        self.timerx = torch.zeros(self.num_drones)
        self.timery = torch.zeros(self.num_drones)
        self.timerz = torch.zeros(self.num_drones)

    def update(self, t, position):
        """
        t := scalar float time
        position := tensor of shape (num_drones, 3), unused
        Returns:
            wind := tensor of shape (num_drones, 3)
        """
        if not self.timers_initialized:
            self.timerx[:] = t
            self.timery[:] = t
            self.timerz[:] = t
            self.timers_initialized = True

        # Update indices if needed
        for i in range(self.num_drones):
            # X-axis
            if (t - self.timerx[i]) >= self.durations[i, 0]:
                if self.random_flag:
                    self.xids[i] = random.randint(0, len(self.wx_ladders[i]) - 1)
                else:
                    self.xids[i] = (self.xids[i] + 1) % len(self.wx_ladders[i])
                self.timerx[i] = t

            # Y-axis
            if (t - self.timery[i]) >= self.durations[i, 1]:
                if self.random_flag:
                    self.yids[i] = random.randint(0, len(self.wy_ladders[i]) - 1)
                else:
                    self.yids[i] = (self.yids[i] + 1) % len(self.wy_ladders[i])
                self.timery[i] = t

            # Z-axis
            if (t - self.timerz[i]) >= self.durations[i, 2]:
                if self.random_flag:
                    self.zids[i] = random.randint(0, len(self.wz_ladders[i]) - 1)
                else:
                    self.zids[i] = (self.zids[i] + 1) % len(self.wz_ladders[i])
                self.timerz[i] = t

        # Assemble wind vectors
        wx = torch.stack([self.wx_ladders[i][self.xids[i]] for i in range(self.num_drones)])
        wy = torch.stack([self.wy_ladders[i][self.yids[i]] for i in range(self.num_drones)])
        wz = torch.stack([self.wz_ladders[i][self.zids[i]] for i in range(self.num_drones)])

        return torch.stack([wx, wy, wz], dim=1)

if __name__ == "__main__":

    # ----- NoWind -----
    print("\n--- NoWind ---")
    wind = NoWind()
    print(wind.update(0, np.array([0, 0, 0])))

    print("\n--- BatchedNoWind ---")
    batched_wind = BatchedNoWind(num_drones=3)
    print(batched_wind.update(0, torch.zeros(3, 3)))

    # ----- ConstantWind -----
    print("\n--- ConstantWind ---")
    wind = ConstantWind(wx=1, wy=1, wz=1)
    print(wind.update(0, np.array([0, 0, 0])))

    # ----- SinusoidWind -----
    print("\n--- SinusoidWind ---")
    wind = SinusoidWind(amplitudes=np.array([1, 2, 3]), frequencies=np.array([0.5, 1.0, 1.5]), phase=np.array([0.5, 0.5, 0]))
    print(wind.update(1.0, np.array([0, 0, 0])))

    print("\n--- BatchedSinusoidWind ---")
    B = 3
    batched_wind = BatchedSinusoidWind(
        amplitudes=torch.tensor([[1.0, 2.0, 3.0]] * B),
        frequencies=torch.tensor([[0.5, 1.0, 1.5]] * B),
        phases=torch.tensor([[0.5, 0.5, 0.0]] * B),
    )
    print(batched_wind.update(1.0, torch.zeros(B, 3)))

    # ----- LadderWind -----
    print("\n--- LadderWind ---")
    wind = LadderWind(min=np.array([0, 0, 0]), max=np.array([1, 1, 1]), duration=np.array([0.5, 0.5, 0.5]), Nstep=np.array([5, 5, 5]), random_flag=False)
    print(wind.update(0.0, np.array([0, 0, 0])))
    print(wind.update(0.6, np.array([0, 0, 0])))

    print("\n--- BatchedLadderWind ---")
    B = 3
    batched_wind = BatchedLadderWind(
        min_vals=torch.tensor([[0.0, 0.0, 0.0]] * B),
        max_vals=torch.tensor([[1.0, 1.0, 1.0]] * B),
        durations=torch.tensor([[0.5, 0.5, 0.5]] * B),
        nsteps=torch.tensor([[5, 5, 5]] * B),
        random_flag=False,
    )
    print(batched_wind.update(0.0, torch.zeros(B, 3)))
    print(batched_wind.update(0.6, torch.zeros(B, 3)))


import numpy as np
import torch

class BatchedWind(object):
    """
    Constructs a batch of wind profiles for multiple vehicles. Each vehicle has its own wind profile. 
    This is a brute force way to interface with RotorPy's batched simulation environment. 
    It's likely to be inefficient, but it's a start. 
    If you want to speed things up, you might want to consider implementing a batched/vectorized version of your wind.
    (see default_wind.py for examples of Batched Wind objects)
    """

    def __init__(self, wind_profiles, device='cpu'):
        """
        Inputs: 
            wind_profiles, a list of wind profile objects, one for each vehicle.
        """

        self.wind_profiles = wind_profiles
        self.num_vehicles = len(wind_profiles)

        self.wind_out = torch.zeros(self.num_vehicles, 3, device=device)

    def update(self, t, position):
        """
        Given the present time and position of the vehicles in the world frame, return the
        current wind speed on all three axes for each vehicle.
        """

        for i in range(self.num_vehicles):
            vehicle_wind = self.wind_profiles[i].update(t, position[i])
            self.wind_out[i] = torch.from_numpy(vehicle_wind)

        return self.wind_out


''' 
Test the estimator module.
'''

import inspect
import numpy as np
import glob
import os
import importlib

EXPECTED_STEP_SIGNATURE = {
    'args': ['self', 'ground_truth_state', 'controller_command', 'imu_measurement', 'mocap_measurement'],
    'return_type': 'dict'
}

EXPECTED_RETURN_DICT = {
    'filter_state': np.ndarray,     # The current filter estimate
    'covariance': np.ndarray        # The current covariance matrix
}

# Function to check the signature of the update method
def check_update_signature(class_obj):
    # Ensure the class has an 'update' method
    assert hasattr(class_obj, 'update'), f"{class_obj.__name__} does not have an 'update' method."

    # Get the signature of the 'update' method
    update_method = getattr(class_obj, 'update')

    # Get the signature of the method using inspect
    signature = inspect.signature(update_method)

    # Get the parameter names and return type
    params = list(signature.parameters.keys())  # Get parameter names as a list

    # Check if the parameter names match
    assert params == EXPECTED_STEP_SIGNATURE['args'], f"Expected arguments {EXPECTED_STEP_SIGNATURE['args']} for {class_obj.__name__}, but got {params}"

# Test case to check all estimators adhere to input requirement.
def test_estimators_step_method():
    
    # Import all estimators from the estimators module
    estimator_files = glob.glob(os.path.join(os.path.dirname(__file__), '..', 'rotorpy', 'estimators', '*.py'))
    estimator_objects = []
    print("\n")
    for file in estimator_files:
        estimator_name = os.path.basename(file)[:-3]  # Remove '.py' extension
        if estimator_name != '__init__': # Ignore the __init__.py file
            print(f"Importing estimator classes from {estimator_name}")
            estimator_module = importlib.import_module(f"rotorpy.estimators.{estimator_name}")  # Import the module
            estimator_classes = inspect.getmembers(estimator_module, inspect.isclass)

            # Ignore any dependency classes in the estimators module
            estimator_objects.extend([estimator_class for estimator_class in estimator_classes if estimator_class[1].__module__ == f"rotorpy.estimators.{estimator_name}"])

    # Iterate over all imported classes from the controllers module
    for estimator_class in estimator_objects:
        print(f"Checking {estimator_class[0]} input signatures...")
        if hasattr(estimator_class[1], 'update'):  # Check only classes with an 'update' method
            check_update_signature(estimator_class[1])


''' 
Test the controller module. 
'''

import inspect
import numpy as np
import glob
import os
import importlib

# The expected signature for the `update` method
EXPECTED_UPDATE_SIGNATURE = {
    'args': ['self', 't', 'state', 'flat_output'],  # Expected arguments, 'self' is implicit
    'return_type': 'dict'  # Expected return type (dict)
}

EXPECTED_RETURN_DICT = {
    'cmd_motor_speeds': np.ndarray,  # numpy array of motor speeds (e.g., [motor1_speed, motor2_speed, ...])
    'cmd_motor_thrusts': np.ndarray, # numpy array of motor thrusts (e.g., [motor1_thrust, motor2_thrust, ...])
    'cmd_thrust': float,             # scalar thrust (N)
    'cmd_moment': np.ndarray,        # numpy array of moments for each axis (e.g., [moment_x, moment_y, moment_z])
    'cmd_q': np.ndarray,             # numpy array representing a quaternion [i, j, k, w]
    'cmd_w': np.ndarray,             # numpy array of angular rates [w_x, w_y, w_z]
    'cmd_v': np.ndarray,             # numpy array representing the velocity vector [v_x, v_y, v_z]
    'cmd_acc': np.ndarray            # numpy array representing the acceleration vector [a_x, a_y, a_z]
}

# Function to check the signature of the update method
def check_update_signature(class_obj):
    # Ensure the class has an 'update' method
    assert hasattr(class_obj, 'update'), f"{class_obj.__name__} does not have an 'update' method."

    # Get the signature of the 'update' method
    update_method = getattr(class_obj, 'update')

    # Get the signature of the method using inspect
    signature = inspect.signature(update_method)

    # Get the parameter names and return type
    params = list(signature.parameters.keys())  # Get parameter names as a list

# Test case to check all controllers adhere to input requirement. 
def test_controllers_update_method():

    # Import all controllers from the controllers module
    controller_files = glob.glob(os.path.join(os.path.dirname(__file__), '..', 'rotorpy', 'controllers', '*.py'))
    controller_objects = []
    print("\n")
    for file in controller_files:
        controller_name = os.path.basename(file)[:-3]  # Remove '.py' extension
        if controller_name != '__init__': # Ignore the __init__.py file
            print(f"Importing controller classes from {controller_name}")
            controller_module = importlib.import_module(f"rotorpy.controllers.{controller_name}")  # Import the module
            controller_classes = inspect.getmembers(controller_module, inspect.isclass)            # Get all classes in the module

            # Ignore any dependency classes in the controllers module
            controller_objects.extend([controller_class for controller_class in controller_classes if controller_class[1].__module__ == f"rotorpy.controllers.{controller_name}"])

    # Iterate over all imported classes from the controllers module
    for controller_class in controller_objects:
        print(f"Checking {controller_class[0]} input signatures...")
        if hasattr(controller_class[1], 'update'):  # Check only classes with an 'update' method
            check_update_signature(controller_class[1])

# Specifically test SE3Control class for both input and output signatures. 
def test_se3_control():
    from rotorpy.controllers.quadrotor_control import SE3Control
    from rotorpy.vehicles.crazyflie_params import quad_params

    print("\nTesting SE3Control class")
    # Create an instance of the SE3Control class
    se3_controller = SE3Control(quad_params)

    return_value = se3_controller.update(0.0, {'x': np.zeros(3), 
                                               'v': np.zeros(3), 
                                               'q': np.array([0, 0, 0, 1]), 
                                               'w': np.zeros(3)}, 
                                              {'x': np.zeros(3), 
                                               'x_dot': np.zeros(3),
                                               'x_ddot': np.zeros(3),
                                               'x_dddot': np.zeros(3),
                                               'x_ddddot': np.zeros(3),
                                               'yaw': 0.0, 
                                               'yaw_dot': 0.0, 
                                               'yaw_ddot': 0.0})

    # Check if the return value is a dictionary
    assert isinstance(return_value, dict), f"SE3Control 'update' method must return a dictionary, but got {type(return_value)}"
    print("SE3Control 'update' method returns a dictionary.")

    # Check if the return value has the expected keys and types
    for key, value in EXPECTED_RETURN_DICT.items():
        assert key in return_value, f"Expected key '{key}' in return dictionary"
        assert isinstance(return_value[key], value), f"Expected type {value} for key '{key}', but got {type(return_value[key])}"


import numpy as np
from scipy.spatial.transform import Rotation
import torch

import gymnasium as gym
from gymnasium import spaces

import math

"""
Reward functions for quadrotor tasks. 
"""

def hover_reward(observation, action, weights={'x': 1, 'v': 0.1, 'w': 0, 'u': 1e-5}):
    """
    Rewards hovering at (0, 0, 0). It is a combination of position error, velocity error, body rates, and 
    action reward.
    """

    # Compute the distance to goal
    dist_reward = -weights['x']*np.linalg.norm(observation[0:3])

    # Compute the velocity reward
    vel_reward = -weights['v']*np.linalg.norm(observation[3:6])

    # Compute the angular rate reward
    ang_rate_reward = -weights['w']*np.linalg.norm(observation[10:13])

    # Compute the action reward
    action_reward = -weights['u']*np.linalg.norm(action)

    return dist_reward + vel_reward + action_reward + ang_rate_reward


def hover_reward_positive(observation, action, weights={'x': 1, 'v': 0.1, 'w': 0, 'u': 1e-5}):
    """
    Rewards hovering at (0, 0, 0). It is a combination of position error, velocity error, body rates, and
    action reward. It inverts the various terms to assign a positive reward for approaching hover.
    """

    # Compute the distance to goal
    dist_reward = weights['x']*1/(1+np.linalg.norm(observation[0:3]))

    # Compute the velocity reward
    vel_reward = weights['v']*1/(1+np.linalg.norm(observation[3:6]))

    # Compute the angular rate reward
    ang_rate_reward = weights['w']*1/(1+np.linalg.norm(observation[10:13]))

    # Compute the action reward
    action_reward = weights['u']*1/(1+np.linalg.norm(action))

    return dist_reward + vel_reward + action_reward + ang_rate_reward


def vec_hover_reward(observation, action, weights={'x': 1, 'v': 0.1, 'w': 0, 'u': 1e-5}):
    """
    Rewards hovering at (0, 0, 0). It is a combination of position error, velocity error, body rates, and
    action reward. Computes rewards for each environment. Adds a positive survival reward.
    """

    # Compute the distance to goal
    dist_reward = -weights['x']*np.linalg.norm(observation[...,0:3], axis=-1)

    # Compute the velocity reward
    vel_reward = -weights['v']*np.linalg.norm(observation[...,3:6], axis=-1)

    # Compute the angular rate reward
    ang_rate_reward = -weights['w']*np.linalg.norm(observation[...,10:13], axis=-1)

    # Compute the action reward
    action_reward = -weights['u']*np.linalg.norm(action, axis=-1)

    return dist_reward + vel_reward + action_reward + ang_rate_reward + 2


def vec_hover_reward_positive(observation, action, weights={'x': 1, 'v': 0.1, 'w': 0, 'u': 1e-5}):
    """
    Rewards hovering at (0, 0, 0). It is a combination of position error, velocity error, body rates, and
    action reward. Computes rewards for each environment.
    """

    # distance reward - reward getting closer to 0
    dist_reward = weights['x'] * 1/(1+np.linalg.norm(observation[...,0:3], axis=-1))

    # Compute the velocity reward
    vel_reward = weights['v'] * 1/(1+np.linalg.norm(observation[...,3:6], axis=-1))

    # Compute the angular rate reward
    ang_rate_reward = weights['w']*1/(1+np.linalg.norm(observation[...,10:13], axis=-1))

    # Compute the action reward
    action_reward = weights['u']*1/(1+np.linalg.norm(action, axis=-1))

    return dist_reward + vel_reward + action_reward + ang_rate_reward



import numpy as np
import torch
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import warnings

from rotorpy.world import World
from rotorpy.vehicles.multirotor import BatchedMultirotorParams, BatchedMultirotor
from rotorpy.vehicles.crazyflie_params import quad_params as crazyflie_params
from rotorpy.learning.quadrotor_reward_functions import vec_hover_reward, vec_hover_reward_positive
from rotorpy.utils.shapes import Quadrotor
from rotorpy.learning.learning_utils import crazyflie_randomizations, update_vehicle_params
from rotorpy.trajectories.hover_traj import HoverTraj

import gymnasium as gym
from stable_baselines3.common.vec_env import VecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnvIndices
from gymnasium import spaces

from copy import deepcopy

DEFAULT_RESET_OPTIONS = {'initial_states': 'random', 'pos_bound': 2, 'vel_bound': 0,
                         "params": "fixed", 
                         "randomization_ranges": crazyflie_randomizations}

def _minmax_scale(x, min_values, max_values):
    '''
    Scales an array of values from [-1, 1] to [min_values, max_values]
    x: np.ndarray of shape (num_envs,...) or scalar
    min_values: np.ndarray of shape (num_envs,...) or scalar
    max_values: np.ndarray of shape (num_envs,...) or scalar
    '''
    x_scaled = np.clip((x+1)/2 * (max_values - min_values) + min_values, min_values, max_values)
    return x_scaled

class QuadrotorEnv(VecEnv):
    """

    A quadrotor environment for reinforcement learning using Gymnasium.

    Inputs:
        initial_states: the initial state of the quadrotor. E.g.
            {'x': torch.zeros(num_drones,3, device=device).double(),
             'v': torch.zeros(num_drones, 3, device=device).double(),
             'q': torch.tensor([0, 0, 0, 1], device=device).repeat(num_drones, 1).double(),
             'w': torch.zeros(num_drones, 3, device=device).double(),
             'wind': torch.zeros(num_drones, 3, device=device).double(),
             'rotor_speeds': torch.tensor([init_rotor_speed, init_rotor_speed, init_rotor_speed, init_rotor_speed], device=device).repeat(num_drones, 1).double()
             }
        control_mode: the appropriate control abstraction that is used by the controller, options are...
                                    'cmd_motor_speeds': the controller directly commands motor speeds.
                                    'cmd_motor_thrusts': the controller commands forces for each rotor.
                                    'cmd_ctbr': the controller commands a collective thrsut and body rates.
                                    'cmd_ctbm': the controller commands a collective thrust and moments on the x/y/z body axes
                                    'cmd_vel': the controller commands a velocity vector in the body frame.
        reward_fn: the reward function, default to hover, but the user can pass in any function that is used as a reward.
        quad_params: the parameters for the quadrotor.
        max_time: the maximum time of the session. After this time, the session will exit.
        world: the world for the quadrotor to operate within.
        sim_rate: the simulation rate (in Hz), i.e. the timestep.
        aero: boolean, determines whether or not aerodynamic wrenches are computed.
        render_mode: render the quadrotor.
        render_fps: rendering frames per second, lower this for faster visualization.
        ax: for plotting purposes, you can supply an axis object that the quadrotor will visualize on.
        color: choose the color of the quadrotor. If none, it will randomly select a color.
    """
    metadata = {"render_modes": ["None", "3D", "console", "debug"],
                "render_fps": 30,
                "num_quads_to_render": 10,
                "control_modes": ['cmd_motor_speeds', 'cmd_motor_thrusts', 'cmd_ctbr', 'cmd_ctbm', 'cmd_vel', 'cmd_ctatt']}

    def __init__(self,
                 num_envs,                              # Number of quadrotor environments to spawn.
                 initial_states,                        # The initial states of the drones.
                 control_mode = 'cmd_vel',              # Control abstraction, see metadata["control_modes"] for a list.
                 reward_fn = vec_hover_reward,          # Reward function, must output the same dim as the number of drones. 
                 quad_params = crazyflie_params,        # Vehicle params for the quadrotor environment. Can be BatchedMultirotorParams. 
                 device = torch.device('cpu'),          # Device to load environment onto. 
                 max_time = 10,                         # Maximum time to run the simulation for in a single session.
                 wind_profile = None,                   # wind profile object, if none is supplied it will choose no wind.
                 world        = None,                   # The world object
                 sim_rate = 100,                        # The update frequency of the simulator in Hz
                 aero = True,                           # Whether or not aerodynamic wrenches are computed.
                 render_mode = "None",                  # The rendering mode
                 render_fps = 30,                       # The rendering frames per second. Lower this for faster visualization.
                 fig = None,                            # Figure for rendering. Optional.
                 ax = None,                             # Axis for rendering. Optional.
                 color = None,                          # The color of the quadrotor.
                 reset_options = DEFAULT_RESET_OPTIONS
                 ):

        self.num_envs = num_envs
        self.device = device

        # Number of environments to render when render is called.
        self.num_quads_to_render = self.metadata['num_quads_to_render']

        # Initial state is a dict of initial states for the quadrotor. Assert that all the keys are present and that they have the right shape. 
        assert all(key in initial_states for key in ('x', 'v', 'q', 'w', 'wind', 'rotor_speeds'))
        for key in initial_states.keys():
            if not isinstance(initial_states[key], torch.Tensor): # Ensure each element is a tensor. 
                initial_states[key] = torch.from_numpy(initial_states[key])
            if initial_states[key].ndim == 1: # Add batch dimension if missing. 
                initial_states[key] = initial_states[key].unsqueeze(0)
            initial_states[key] = initial_states[key].to(device)

        self.initial_states = deepcopy(initial_states)
        self.vehicle_states = deepcopy(initial_states)

        # Construct BatchedMultirotorParams from quad_params. 
        if type(quad_params) == dict: # if only one quadrotor param config is given, convert it to BatchedMultirotorParams for legacy support. 
            self.quad_params = BatchedMultirotorParams([quad_params for _ in range(num_envs)], num_envs, device=self.device)
        elif isinstance(quad_params, BatchedMultirotorParams):
            self.quad_params = quad_params
        else:
            raise ValueError("quad_params must be either a dictionary (if all envs have the same params) or of type BatchedMultirotorParams")

        assert control_mode in self.metadata["control_modes"]  # Don't accept improper control modes
        self.control_mode = control_mode

        self.sim_rate = sim_rate
        self.t_step = 1/self.sim_rate
        self.reward_fn = reward_fn
        self.metadata["render_fps"] = render_fps

        self.quadrotors = BatchedMultirotor(batched_params=self.quad_params,
                                           num_drones = self.num_envs,
                                           initial_states=initial_states,
                                           device=self.device,
                                           control_abstraction=control_mode,
                                           aero=aero,
                                           integrator="rk4")
        self.t = np.zeros(self.num_envs)
        self.max_time = max_time

        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(13,), dtype=np.float32)

        # Stable Baselines expects the action/observation space to be the same as a single env (ie. not batched)
        if self.control_mode == 'cmd_vel':
            self.action_space = spaces.Box(low = -1, high = 1, shape = (3,), dtype=np.float32)
        else:
            self.action_space = spaces.Box(low = -1, high = 1, shape = (4,), dtype=np.float32)

        # Get the minimum and maximum rotor speeds directly from quad_params. 
        self.rotor_speed_max = self.quad_params.rotor_speed_max.cpu().numpy()
        self.rotor_speed_min = self.quad_params.rotor_speed_min.cpu().numpy()

        # Compute the min/max thrust by assuming the rotor is spinning at min/max speed. (also generalizes to bidirectional rotors)
        self.max_thrust = self.quad_params.k_eta.cpu().numpy() * self.rotor_speed_max**2
        self.min_thrust = self.quad_params.k_eta.cpu().numpy() * self.rotor_speed_min**2

        # Find the maximum moment on each axis, N-m
        self.max_roll_moment = np.array([self.max_thrust[i] * np.abs(self.quad_params.rotor_pos[i]['r1'][1]) for i in range(self.num_envs)])
        self.max_pitch_moment = np.array([self.max_thrust[i] * np.abs(self.quad_params.rotor_pos[i]['r1'][0]) for i in range(self.num_envs)])
        self.max_yaw_moment = self.quad_params.k_m.cpu().numpy() * self.rotor_speed_max**2

        # Set the maximum body rate on each axis (this is hand selected), rad/s
        self.max_roll_br = 7.0
        self.max_pitch_br = 7.0 
        self.max_yaw_br = 3.0

        self.max_vel = 4/np.sqrt(4)   # Selected so that at most the max speed is 4 m/s
        self.rotor_speed_order_mag = np.floor(np.log10(self.rotor_speed_max))
        if render_mode.lower() == 'human':
            render_mode = '3D'
        self.render_mode = [render_mode] * self.num_envs
        self.reward = np.zeros(self.num_envs)

        if world is None:
            # If no world is specified, assume that it means that the intended world is free space.
            wbound = 4
            self.world = World.empty((-wbound, wbound, -wbound,
                                      wbound, -wbound, wbound))
        else:
            self.world = world

        if wind_profile is None:
            # If wind is not specified, default to no wind.
            from rotorpy.wind.default_winds import BatchedNoWind
            self.wind_profile = BatchedNoWind(num_drones=num_envs)
        else:
            self.wind_profile = wind_profile

        if render_mode == '3D': 
            warnings.warn("3D Rendering in Vectorized Environment behaves strangely during training.")
            if fig is None and ax is None:
                self.fig = plt.figure('Visualization')
                self.ax = self.fig.add_subplot(projection='3d')
            else:
                self.fig = fig
                self.ax = ax
            if color is None:
                colors = list(mcolors.CSS4_COLORS)
            else:
                colors = [color]
            self.quad_objs = [Quadrotor(self.ax, wind=True, color=np.random.choice(colors), wind_scale_factor=5) for _ in range(min(self.num_quads_to_render, self.num_envs))]
            self.world_artists = None
            self.title_artist = self.ax.set_title('t = {}'.format(self.t))

        self.default_rotor_speed = torch.sqrt(self.quad_params.mass*self.quad_params.g/(self.quad_params.num_rotors*self.quad_params.k_eta))
        self.reset_options = dict(reset_options)

        self.debug_states = np.zeros((min(self.num_envs, 5), int(self.max_time * self.sim_rate), 3))
        self.counts = np.zeros(min(self.num_envs, 5))
        self.cum_rewards = np.zeros(self.num_envs)

    def close(self):
        pass

    def reset_idx(self, env_idx, options):
        """ 
        Resets the i'th quadrotor in the environment. 
        """ 
        
        if options['initial_states'] == 'random':
            pos = torch.rand(3, device=self.device, dtype=torch.float64) * 2 * options['pos_bound'] - options['pos_bound']
            vel = torch.rand(3, device=self.device, dtype=torch.float64) * 2 * options['vel_bound'] - options['vel_bound']
            self.vehicle_states['x'][env_idx] = pos.double()
            self.vehicle_states['v'][env_idx] = vel.double()
            self.vehicle_states['q'][env_idx] = torch.tensor([0, 0, 0, 1], device=self.device).double()
            self.vehicle_states['w'][env_idx] = torch.zeros(3, device=self.device).double()
            self.vehicle_states['wind'][env_idx] = torch.zeros(3, device=self.device).double()
            self.vehicle_states['wind'][env_idx] = torch.zeros(3, device=self.device).double()
            self.vehicle_states['rotor_speeds'][env_idx] = torch.ones(4, device=self.device).double() * self.default_rotor_speed[env_idx]
        elif options['initial_states'] == 'deterministic':
            self.vehicle_states = self.initial_states
        elif isinstance(options['initial_states'], dict):
            # Ensure the correct keys are in dict.
            assert all(key in options['initial_states'] for key in ('x', 'v', 'q', 'w', 'wind', 'rotor_speeds'))
            for key in options['initial_states'].keys():
                self.vehicle_states[key][env_idx] = self.initial_states[key][env_idx].double().to(self.device)
        else:
            raise ValueError(
                "You must either specify 'random', 'deterministic', or provide a dict containing your desired initial state.")

        self.t[env_idx] = 0.0
        self.reward[env_idx] = 0.0
        self.cum_rewards[env_idx] = 0.0

        if options["params"] == "random":
            update_vehicle_params(env_idx, options["randomization_ranges"], self.quad_params)

            # update necessary scaling params of environment
            self.min_thrust[env_idx] = self.quad_params.k_eta[env_idx].cpu().numpy() * self.rotor_speed_min[env_idx]**2
            self.max_thrust[env_idx] = self.quad_params.k_eta[env_idx].cpu().numpy() * self.rotor_speed_max[env_idx]**2
            self.max_roll_moment[env_idx] = self.max_thrust[env_idx] * np.abs(self.quad_params.rotor_pos[env_idx]['r1'][1])
            self.max_pitch_moment[env_idx] = self.max_thrust[env_idx] * np.abs(self.quad_params.rotor_pos[env_idx]['r1'][0])
            self.max_yaw_moment = self.quad_params.k_m[env_idx].cpu().numpy() * self.rotor_speed_max**2

        if env_idx < min(self.num_envs, 5) and self.render_mode[0] == "debug":
            if self.counts[env_idx] % 10 == 0 and self.counts[env_idx] != 0:
                print(f"Saving plot for {env_idx}")
                fig, ax  = plt.subplots(1, 3)
                ax[0].plot(self.debug_states[env_idx,:,0])
                ax[1].plot(self.debug_states[env_idx,:,1])
                ax[2].plot(self.debug_states[env_idx,:,2])
                plt.savefig(f"debug_states_{env_idx}_{self.counts[env_idx]}.png")
            self.debug_states[env_idx] = 0
            self.counts[env_idx] += 1

    def reset(self, seed=None, options=None):
        """
        Reset the environment
        Inputs:
            seed: the seed for any random number generation, mostly for reproducibility.
            options: dictionary for misc options for resetting the scene.
                        'initial_states': determines how to set the quadrotor again. Options are...
                            'random': will randomly select the state of the quadrotor.
                            'deterministic': will set the state to the initial state selected by the user when creating
                                            the quadrotor environment (usually hover).
                            the user can also specify the state itself as a dictionary... e.g.
                                reset(options={'initial_states':
                                    {'x': torch.zeros(num_drones,3, device=device).double(),
                                     'v': torch.zeros(num_drones, 3, device=device).double(),
                                     'q': torch.tensor([0, 0, 0, 1], device=device).repeat(num_drones, 1).double(),
                                     'w': torch.zeros(num_drones, 3, device=device).double(),
                                     'wind': torch.zeros(num_drones, 3, device=device).double(),
                                     'rotor_speeds': torch.tensor([init_rotor_speed, init_rotor_speed, init_rotor_speed, init_rotor_speed], device=device).repeat(num_drones, 1).double()
                                     }
                        'pos_bound': the min/max position region for random placement.
                        'vel_bound': the min/max velocity region for random placement

        """
        options = self.reset_options
        # If any options are not specified, set them to default values.
        if 'pos_bound' not in options:
            options['pos_bound'] = 2
        if 'vel_bound' not in options:
            options['vel_bound'] = 0
        if 'initial_states' not in options:
            options['initial_states'] = 'random'

        # Assert that the bounds are greater than or equal to 0.
        assert options['pos_bound'] >= 0 and options['vel_bound'] >= 0, "Bounds must be greater than or equal to 0."

        # Reset the gym environment
        if seed is not None:
            np.random.seed(seed)
            torch.manual_seed(seed)

        for i in range(self.num_envs):
            self.reset_idx(i, options)

        # Reset the time
        self.t = np.zeros(self.num_envs)

        # Reset the rewards
        self.reward = np.zeros(self.num_envs)

        # Now get observation and info using the new state
        observation = self._get_obs()
        self.reset_infos = {}

        self.render()

        return observation

    def step(self, action):
        """
        Step the quadrotor dynamics forward by one step based on the policy action. 
        Inputs:
            action: The action is a 4x1 vector which depends on the control abstraction: 

            if control_mode == 'cmd_vel':
                action[0] (-1,1) := commanded velocity in x direction (will be rescaled to m/s)
                action[1] (-1,1) := commanded velocity in y direction (will be rescaled to m/s)
                action[2] (-1,1) := commanded velocity in z direction (will be rescaled to m/s)
            if control_mode == 'cmd_ctbr':
                action[0] (-1,1) := the thrust command (will be rescaled to Newtons)
                action[1] (-1,1) := the roll body rate (will be rescaled to rad/s)
                action[2] (-1,1) := the pitch body rate (will be rescaled to rad/s)
                action[3] (-1,1) := the yaw body rate (will be rescaled to rad/s)
            if control_mode == 'cmd_ctbm':
                action[0] (-1,1) := the thrust command (will be rescaled to Newtons)
                action[1] (-1,1) := the roll moment (will be rescaled to Newton-meters)
                action[2] (-1,1) := the pitch moment (will be rescaled to Newton-meters)
                action[3] (-1,1) := the yaw moment (will be rescaled to Newton-meters)
            if control_mode == 'cmd_motor_speeds':
                action[0] (-1,1) := motor 1 speed (will be rescaled to rad/s)
                action[1] (-1,1) := motor 2 speed (will be rescaled to rad/s)
                action[2] (-1,1) := motor 3 speed (will be rescaled to rad/s)
                action[3] (-1,1) := motor 4 speed (will be rescaled to rad/s)
            if control_mode == 'cmd_motor_forces':
                action[0] (-1,1) := motor 1 force (will be rescaled to Newtons)
                action[1] (-1,1) := motor 2 force (will be rescaled to Newtons)
                action[2] (-1,1) := motor 3 force (will be rescaled to Newtons)
                action[3] (-1,1) := motor 4 force (will be rescaled to Newtons)

        """

        # First rescale the action and get the appropriate control dictionary given the control mode.
        self.control_dict = self.rescale_action(action)
        for key in self.control_dict.keys():
            self.control_dict[key] = torch.from_numpy(self.control_dict[key]).to(self.device).double()

        # Now update the wind state using the wind profile
        self.vehicle_states['wind'] = self.wind_profile.update(self.t, self.vehicle_states['x'])

        # Last perform forward integration using the commanded motor speed and the current state
        self.vehicle_states = self.quadrotors.step(self.vehicle_states, self.control_dict, self.t_step)
        pre_termination_obs = self._get_obs()

        # Update t by t_step
        self.t += self.t_step

        # Check for safety
        oob = self._is_out_of_bounds()

        # Determine whether or not the session should terminate.
        time_done = self.t >= self.max_time
        dones = np.logical_or(oob, time_done)
        if self.render_mode[0] == "debug":
            print(f"time {self.t}, dones = {dones}")

        # Now compute the reward based on the current state
        self.reward = self._get_reward(pre_termination_obs, action)
        self.cum_rewards += self.reward

        # Finally get info
        infos = [{} for _ in range(self.num_envs)]
        for env_idx in range(self.num_envs):
            if time_done[env_idx]:
                infos[env_idx]["terminal_observation"] = pre_termination_obs
                infos[env_idx]["TimeLimit.truncated"] = True
                self.reset_idx(env_idx, self.reset_options)
            elif oob[env_idx]:
                infos[env_idx]["terminal_observation"] = pre_termination_obs
                infos[env_idx]["TimeLimit.truncated"] = False
                self.reset_idx(env_idx, self.reset_options)

        self.render()

        observation = self._get_obs()
        for i in range(min(self.num_envs, 5)):
            self.debug_states[i,int(self.t[i]/self.t_step)] = self.vehicle_states['x'][i].cpu().numpy()

        return (observation, self.reward, dones, infos)

    def rescale_action(self, action):
        """
        Rescales the action to within the control limits and then assigns the appropriate dictionary, and converts to pytorch tensors.
        """

        control_dict = {}

        if self.control_mode == 'cmd_ctbm':
            # Scale action[0] to (0,1) and then scale to the max thrust
            cmd_thrust = _minmax_scale(action[...,0].reshape(-1, 1), self.quad_params.num_rotors * self.min_thrust, self.quad_params.num_rotors * self.max_thrust)

            # Scale the moments
            cmd_roll_moment = _minmax_scale(action[...,1].reshape(-1, 1), -self.max_roll_moment, self.max_roll_moment)
            cmd_pitch_moment = _minmax_scale(action[...,2].reshape(-1, 1), -self.max_pitch_moment, self.max_pitch_moment)
            cmd_yaw_moment = _minmax_scale(action[...,3].reshape(-1, 1), -self.max_yaw_moment, self.max_yaw_moment)

            control_dict['cmd_thrust'] = cmd_thrust
            control_dict['cmd_moment'] = np.hstack([cmd_roll_moment, cmd_pitch_moment, cmd_yaw_moment])

        elif self.control_mode == 'cmd_ctbr':
            # Scale action to min and max thrust.
            cmd_thrust = _minmax_scale(action[...,0].reshape(-1, 1), self.quad_params.num_rotors * self.min_thrust, self.quad_params.num_rotors * self.max_thrust)

            # Scale the body rates.
            cmd_roll_br = _minmax_scale(action[...,1], -self.max_roll_br, self.max_roll_br).reshape(-1, 1)
            cmd_pitch_br = _minmax_scale(action[...,2], -self.max_pitch_br, self.max_pitch_br).reshape(-1, 1)
            cmd_yaw_br = _minmax_scale(action[...,3], -self.max_yaw_br, self.max_yaw_br).reshape(-1, 1)
            control_dict['cmd_thrust'] = cmd_thrust
            control_dict['cmd_w'] = np.hstack([cmd_roll_br, cmd_pitch_br, cmd_yaw_br])

        elif self.control_mode == 'cmd_motor_speeds':
            # Scale the action to min and max motor speeds.
            control_dict['cmd_motor_speeds'] = _minmax_scale(action, self.rotor_speed_min, self.rotor_speed_max)

        elif self.control_mode == 'cmd_motor_thrusts':
            # Scale the action to min and max rotor thrusts.
            control_dict['cmd_motor_thrusts'] = _minmax_scale(action, self.min_thrust, self.max_thrust)

        elif self.control_mode == 'cmd_vel':
            # Scale the velcoity to min and max values.
            control_dict['cmd_v'] = _minmax_scale(action, -self.max_vel, self.max_vel)
        elif self.control_mode == "cmd_ctatt":
            cmd_thrust = _minmax_scale(action[...,0].reshape(-1, 1), self.quad_params.num_rotors * self.min_thrust, self.quad_params.num_rotors * self.max_thrust)
            control_dict["cmd_thrust"] = cmd_thrust
            control_dict["cmd_q"] = Rotation.from_euler("xyz", _minmax_scale(action[...,1:4].reshape(-1, 3), -np.pi, np.pi)).as_quat()
        return control_dict

    def _get_reward(self, observation, action):
        return self.reward_fn(observation, action)

    def _is_out_of_bounds(self):
        se = torch.any(torch.abs(self.vehicle_states['v']) > 100, dim=-1)
        if self.render_mode[0] == "debug":
            print(se)
        se = torch.logical_or(se, torch.any(torch.abs(self.vehicle_states['w']) > 100, dim=-1))
        if self.render_mode[0] == "debug":
            print(se)
        se = torch.logical_or(se,
                              torch.logical_or(
                                  self.vehicle_states['x'][:,0] < self.world.world['bounds']['extents'][0],
                                  self.vehicle_states['x'][:,0] > self.world.world['bounds']['extents'][1]
                              ))
        if self.render_mode[0] == "debug":
            print(se)
        se = torch.logical_or(se,
                              torch.logical_or(
                                  self.vehicle_states['x'][:,1] < self.world.world['bounds']['extents'][2],
                                  self.vehicle_states['x'][:,1] > self.world.world['bounds']['extents'][3]
                              ))
        if self.render_mode[0] == "debug":
            print(se)
        se = torch.logical_or(se,
                              torch.logical_or(
                                  self.vehicle_states['x'][:,2] < self.world.world['bounds']['extents'][4],
                                  self.vehicle_states['x'][:,2] > self.world.world['bounds']['extents'][5]
                              ))
        if self.render_mode[0] == "debug":
            print(se)
            print("----------")
        return se.cpu().numpy()

    def _get_obs(self):
        # Concatenate all the state variables into a single vector
        state_vec = torch.cat([self.vehicle_states['x'],
                               self.vehicle_states['v'],
                               self.vehicle_states['q'],
                               self.vehicle_states['w']], dim=-1)
        return state_vec.float().cpu().numpy()

    def seed(self, seed=None):
        np.random.seed(seed)
        torch.manual_seed(seed)

    def render(self, mode=None):
        if self.render_mode[0] == '3D':
            # only plot a maximum of num_evs quads, so the plot is still interpretable.
            for env_idx in range(min(self.num_envs, self.num_quads_to_render)):
                self._plot_quad(env_idx)
        elif self.render_mode[0] == 'console':
            for env_idx in range(min(self.num_envs, 10)):
                self._print_quad(env_idx)

    def _print_quad(self, env_idx):
        # Print the first quadrotor. 
        print("env_idx: %d Time: %3.2f \t Position: (%3.2f, %3.2f, %3.2f) \t Reward: %3.2f" % (env_idx,
            self.t[env_idx], self.vehicle_states['x'][env_idx][0], self.vehicle_states['x'][env_idx][1], self.vehicle_states['x'][env_idx][2], self.reward[env_idx]))

    def _plot_quad(self, env_idx):
        if abs(self.t[env_idx] / (1 / self.metadata['render_fps']) - round(self.t[env_idx] / (1 / self.metadata['render_fps']))) > 5e-2:
            self.rendering = False  # Set rendering bool to false.
            return

        self.rendering = True  # Set rendering bool to true.

        plot_position = deepcopy(self.vehicle_states['x'][env_idx].cpu().numpy())
        plot_rotation = Rotation.from_quat(self.vehicle_states['q'][env_idx].cpu().numpy()).as_matrix()
        plot_wind = deepcopy(self.vehicle_states['wind'][env_idx].cpu().numpy())

        if self.world_artists is None and not ('x' in self.ax.get_xlabel()):
            self.world_artists = self.world.draw(self.ax)
            self.ax.plot(0, 0, 0, 'go')

        self.quad_objs[env_idx].transform(position=plot_position, rotation=plot_rotation, wind=plot_wind)
        self.title_artist.set_text('t = {:.2f}'.format(self.t[env_idx]))

        plt.pause(1e-9)

        return
        
    # Methods to comply with Stable Baselines API
    def set_attr(self, attr_name, value, indices = None):
        pass

    def get_attr(self, attr_name, indices = None):
        return getattr(self, attr_name)

    def env_is_wrapped(self, wrapper_class, indices = None):
        return [False for i in range(self.num_envs)]

    def env_method(self, method_name, *method_args, indices = None, **method_kwargs):
        pass

    def step_async(self, actions):
        self.async_action = actions

    def step_wait(self):
        results = self.step(self.async_action)
        self.async_action = None
        return results

def make_default_vec_env(num_envs, quad_params, control_mode, device, **kwargs):
    num_drones = num_envs
    init_rotor_speed = 1788.53
    x0 = {'x': torch.zeros(num_drones,3, device=device).double(),
          'v': torch.zeros(num_drones, 3, device=device).double(),
          'q': torch.tensor([0, 0, 0, 1], device=device).repeat(num_drones, 1).double(),
          'w': torch.zeros(num_drones, 3, device=device).double(),
          'wind': torch.zeros(num_drones, 3, device=device).double(),
          'rotor_speeds': torch.tensor([init_rotor_speed, init_rotor_speed, init_rotor_speed, init_rotor_speed], device=device).repeat(num_drones, 1).double()}
    return QuadrotorEnv(num_envs, initial_states=x0, quad_params=quad_params, max_time=5, control_mode=control_mode, device=device, **kwargs)


import numpy as np
from scipy.spatial.transform import Rotation
import torch
import copy

class Imu:
    """
    Simulated IMU measurement given
      1) quadrotor's ground truth state and acceleration, and
      2) IMU's pose in quadrotor body frame.

      CREDIT:
      Partial implementation from Yu-Ming Chen and MEAM 620 Sp2022 Teaching Staff
      Finishing touches added by Alexander Spinos, checked by Spencer Folk.
    """
    def __init__(self, accelerometer_params={'initial_bias': np.array([0.,0.,0.]), # m/s^2
                                             'noise_density': (0.038**2)*np.ones(3,), # m/s^2 / sqrt(Hz)
                                             'random_walk': np.zeros(3,) # m/s^2 * sqrt(Hz)
                                             }, 
                gyroscope_params={'initial_bias': np.array([0.,0.,0.]), # m/s^2
                                  'noise_density': (0.01**2)*np.ones(3,), # rad/s / sqrt(Hz)
                                  'random_walk': np.zeros(3,) # rad/s * sqrt(Hz)
                                  }, 
                R_BS = np.eye(3),
                p_BS = np.zeros(3,),
                sampling_rate=500, 
                gravity_vector=np.array([0,0,-9.81])):
        """
        Parameters:
            R_BS, the rotation matrix of sensor frame S in body frame B
            p_BS, the position vector from frame B's origin to frame S's origin, expressed in frame B
            accelerometer_params, a dict with keys:
                initial_bias, accelerometer contant bias,    [ m / s^2 ]
                noise_density, accelerometer "white noise",  [ m / s^2 / sqrt(Hz) ]
                random_walk, accelerometer bias diffusion,   [ m / s^2 * sqrt(Hz) ]
            gyroscope_params, a dict with keys:
                initial_bias, gyro contant bias,             [ m / s^2 ]
                noise_density, gyro "white noise",           [ rad / s / sqrt(Hz) ]
                random_walk, gyro bias diffusion,            [ rad / s * sqrt(Hz) ]
            sampling_rate, the sampling rate of the sensor, Hz (1/s)
            gravity_vector, the gravitational vector in world frame (should be ~ [0, 0 , -9.81])
        """

        # A few checks
        if type(R_BS) != np.ndarray:
            raise TypeError("R_BS's type is not numpy.ndarray")
        if type(p_BS) != np.ndarray:
            raise TypeError("p_BS's type is not numpy.ndarray")
        if type(gravity_vector) != np.ndarray:
            raise TypeError("gravity_vector's type is not numpy.ndarray")
        if R_BS.shape != (3, 3):
            raise ValueError("R_BS's size is not (3, 3)")
        if p_BS.shape != (3,):
            raise ValueError("p_BS's size is not (3,)")
        if gravity_vector.shape != (3,):
            raise ValueError("gravity_vector's size is not (3,)")

        self.R_BS = R_BS
        self.p_BS = p_BS
        self.rate_scale = np.sqrt(sampling_rate/2)
        self.gravity_vector = gravity_vector

        self.accel_variance = accelerometer_params['noise_density'].astype('float64')
        self.accel_random_walk = accelerometer_params['random_walk'].astype('float64')
        self.accel_bias = accelerometer_params['initial_bias'].astype('float64')
        self.gyro_variance = gyroscope_params['noise_density'].astype('float64')
        self.gyro_random_walk = gyroscope_params['random_walk'].astype('float64')
        self.gyro_bias = gyroscope_params['initial_bias'].astype('float64')

    def bias_step(self):
        """Simulate bias drift"""

        self.accel_bias += np.random.normal(scale=self.accel_random_walk) / self.rate_scale
        self.gyro_bias += np.random.normal(scale=self.gyro_random_walk) / self.rate_scale

        return

    def measurement(self, state, acceleration, with_noise=True):
        """
        Computes and returns the IMU measurement at a time step.

        Inputs:
            state, a dict describing the state with keys
                x, position, m, shape=(3,)
                v, linear velocity, m/s, shape=(3,)
                q, quaternion [i,j,k,w], shape=(4,)
                w, angular velocity (in LOCAL frame!), rad/s, shape=(3,)
            acceleration, a dict describing the acceleration with keys
                vdot, quadrotor's linear acceleration expressed in world frame, m/s^2, shape=(3,)
                wdot, quadrotor's angular acceleration expressed in world frame, rad/s^2, shape=(3,)
        Outputs:
            observation, a dict describing the IMU measurement with keys
                accel, simulated accelerometer measurement, m/s^2, shape=(3,)
                gyro, simulated gyroscope measurement, rad/s^2, shape=(3,)
        """
        q_WB = state['q']
        w_WB = state['w']
        alpha_WB_W = acceleration['wdot']
        a_WB_W = acceleration['vdot']

        # Rotation matrix of the body frame B in world frame W
        R_WB = Rotation.from_quat(q_WB).as_matrix()

        # Sensor position in body frame expressed in world coordinates
        p_BS_W = R_WB @ self.p_BS

        # Linear acceleration of point S (the imu) expressed in world coordinates W.
        a_WS_W = a_WB_W + np.cross(alpha_WB_W, p_BS_W) + np.cross(w_WB, np.cross(w_WB, p_BS_W))

        # Rotation from world to imu: R_SW = R_SB * R_BW
        R_SW = self.R_BS.T @ R_WB.T

        # Rotate to local frame
        accelerometer_measurement = R_SW @ (a_WS_W - self.gravity_vector)
        gyroscope_measurement = copy.deepcopy(w_WB).astype(float)

        # Add the bias drift (default 0)
        self.bias_step()

        # Add biases and noises
        accelerometer_measurement += self.accel_bias
        gyroscope_measurement += self.gyro_bias
        if with_noise:
            accelerometer_measurement += self.rate_scale * np.random.normal(scale=np.abs(self.accel_variance))
            gyroscope_measurement +=  self.rate_scale * np.random.normal(scale=np.abs(self.gyro_variance))

        return {'accel': accelerometer_measurement, 'gyro': gyroscope_measurement}

class BatchedImu:
    def __init__(self, num_drones,
                 accelerometer_params=None,
                 gyroscope_params=None,
                 R_BS=None,
                 p_BS=None,
                 sampling_rate=500,
                 gravity_vector=None,
                 device='cpu'):
        self.device = device
        self.num_drones = num_drones
        self.sampling_rate = sampling_rate
        self.rate_scale = (sampling_rate / 2) ** 0.5

        self.R_BS = torch.eye(3, device=device).double() if R_BS is None else R_BS.double().to(device)
        self.p_BS = torch.zeros(3, device=device).double() if p_BS is None else p_BS.double().to(device)
        self.gravity_vector = torch.tensor([0, 0, -9.81], device=device, dtype=torch.double) if gravity_vector is None else gravity_vector.double().to(device)

        accel_params = accelerometer_params or {
            'initial_bias': torch.zeros(3, device=device).double(),
            'noise_density': (0.38 ** 2) * torch.ones(3, device=device).double(),
            'random_walk': torch.zeros(3, device=device).double()
        }

        gyro_params = gyroscope_params or {
            'initial_bias': torch.zeros(3, device=device).double(),
            'noise_density': (0.01 ** 2) * torch.ones(3, device=device).double(),
            'random_walk': torch.zeros(3, device=device).double()
        }

        # Repeat biases for each drone
        self.accel_bias = torch.Tensor(accel_params['initial_bias'][np.newaxis].repeat(num_drones, 1)).double()
        self.gyro_bias = torch.Tensor(gyro_params['initial_bias'][np.newaxis].repeat(num_drones, 1)).double()
        self.accel_noise = torch.Tensor(accel_params['noise_density']).double()
        self.gyro_noise = torch.Tensor(gyro_params['noise_density']).double()
        self.accel_random_walk = torch.Tensor(accel_params['random_walk']).double()
        self.gyro_random_walk = torch.Tensor(gyro_params['random_walk']).double()

    def bias_step(self):
        self.accel_bias += torch.randn_like(self.accel_bias) * (self.accel_random_walk / self.rate_scale)
        self.gyro_bias += torch.randn_like(self.gyro_bias) * (self.gyro_random_walk / self.rate_scale)

    def measurement(self, state, acceleration, idxs=None, with_noise=True):
        """
        state['x'], ['v'], ['w']: (num_drones, 3)
        state['q']: (num_drones, 4) (xyzw format)
        acceleration['vdot'], ['wdot']: (B, 3)
        idxs: list of drones in the batch for which to compute IMU measurements.
        """
        if idxs is None:
            idxs = [i for i in range(self.num_drones)]

        q_WB = state['q'][idxs]  # (num_drones, 4)
        w_WB = state['w'][idxs]
        alpha_WB_W = acceleration['wdot'][idxs]
        a_WB_W = acceleration['vdot'][idxs]

        # Get rotation matrices from quaternions (num_drones, 3, 3)
        R_WB = self.quat_to_rotmat(q_WB)

        # Sensor offset in world frame
        p_BS_W = torch.einsum('bij,j->bi', R_WB, self.p_BS)

        cross_w_p = torch.cross(w_WB, p_BS_W, dim=-1)
        cross_w_w_p = torch.cross(w_WB, cross_w_p, dim=-1)
        cross_alpha_p = torch.cross(alpha_WB_W, p_BS_W, dim=-1)
        a_WS_W = a_WB_W + cross_alpha_p + cross_w_w_p

        R_SW = torch.einsum('ij,bij->bij', self.R_BS.T, R_WB.transpose(1, 2))
        a_WS_S = torch.einsum('bij,bj->bi', R_SW, a_WS_W - self.gravity_vector)

        # Apply bias + noise
        self.bias_step()
        a_meas = a_WS_S + self.accel_bias[idxs]
        w_meas = w_WB + self.gyro_bias[idxs]

        if with_noise:
            a_meas += self.rate_scale * torch.randn_like(a_meas) * self.accel_noise
            w_meas += self.rate_scale * torch.randn_like(w_meas) * self.gyro_noise

        a_meas_out = torch.full((self.num_drones, 3), float("nan"), device=a_meas.device).double()
        w_meas_out = torch.full((self.num_drones, 3), float("nan"), device=w_meas.device).double()
        a_meas_out[idxs] = a_meas
        w_meas_out[idxs] = w_meas

        return {'accel': a_meas_out, 'gyro': w_meas_out}

    @staticmethod
    def quat_to_rotmat(q):
        # Input shape (B, 4) in xyzw
        x, y, z, w = q[:, 0], q[:, 1], q[:, 2], q[:, 3]
        B = q.shape[0]
        R = torch.empty((B, 3, 3), device=q.device, dtype=torch.double)
        R[:, 0, 0] = 1 - 2*(y**2 + z**2)
        R[:, 0, 1] = 2*(x*y - z*w)
        R[:, 0, 2] = 2*(x*z + y*w)
        R[:, 1, 0] = 2*(x*y + z*w)
        R[:, 1, 1] = 1 - 2*(x**2 + z**2)
        R[:, 1, 2] = 2*(y*z - x*w)
        R[:, 2, 0] = 2*(x*z - y*w)
        R[:, 2, 1] = 2*(y*z + x*w)
        R[:, 2, 2] = 1 - 2*(x**2 + y**2)
        return R

if __name__ == "__main__":

    sim_rate = 100
    accelerometer_params = {
        'initial_bias': np.array([0, 0, 0]),  # m/s^2
        'noise_density': (0.38**2) * np.ones(3,),  # m/s^2 / sqrt(Hz)
        'random_walk': np.zeros(3,)  # m/s^2 * sqrt(Hz)
    }
    gyroscope_params = {
        'initial_bias': np.array([0, 0, 0]),  # rad/s
        'noise_density': (0.01**2) * np.ones(3,),  # rad/s / sqrt(Hz)
        'random_walk': np.zeros(3,)  # rad/s * sqrt(Hz)
    }

    imu = Imu(accelerometer_params,
              gyroscope_params,
              p_BS=np.zeros(3,),
              R_BS=np.eye(3),
              sampling_rate=sim_rate)

    state_np = {'x': np.array([0, 0, 0]),
                'v': np.array([0, 0, 0]),
                'q': np.array([0, 0, 0, 1]),
                'w': np.array([0, 0, 0])}

    accel_np = {'vdot': np.array([0, 0, 0]),
                'wdot': np.array([0, 0, 0])}

    meas_single = imu.measurement(state_np, accel_np, with_noise=False)

    print("=== Single Drone IMU Measurement ===")
    print("Accel:", meas_single['accel'])
    print("Gyro: ", meas_single['gyro'])

    num_drones = 10  # batch size
    batched_imu = BatchedImu(num_drones=1,
                              accelerometer_params=accelerometer_params,
                              gyroscope_params=gyroscope_params,
                              p_BS=torch.zeros(3),
                              R_BS=torch.eye(3),
                              sampling_rate=sim_rate)

    state_batch = {
        'x': torch.zeros(num_drones, 3),
        'v': torch.zeros(num_drones, 3),
        'q': torch.tensor([0., 0., 0., 1.]).unsqueeze(0).repeat(num_drones, 1),
        'w': torch.zeros(num_drones, 3)
    }

    accel_batch = {
        'vdot': torch.zeros(num_drones, 3),
        'wdot': torch.zeros(num_drones, 3)
    }

    meas_batch = batched_imu.measurement(state_batch, accel_batch, with_noise=False)

    print("\n=== Batched IMU Measurement ===")
    print("Accel:", meas_batch['accel'].numpy())
    print("Gyro: ", meas_batch['gyro'].numpy())

    print("\n=== Difference ===")
    print("Accel diff:", np.abs(meas_single['accel'] - meas_batch['accel'][0].numpy()))
    print("Gyro diff: ", np.abs(meas_single['gyro'] - meas_batch['gyro'][0].numpy()))


''' 
Test the RotorPy quadrotor gymnasium environment. 
'''

import gymnasium as gym
import numpy as np
import torch
import matplotlib.pyplot as plt
from rotorpy.controllers.quadrotor_control import SE3Control
from rotorpy.vehicles.crazyflie_params import quad_params
from rotorpy.learning.quadrotor_environments import QuadrotorEnv
from rotorpy.learning.quadrotor_reward_functions import hover_reward

def run_gym_environment(control_abstraction, num_drones):

    init_rotor_speed = 1788.53
    x0 = {'x': torch.zeros(num_drones,3).double(),
          'v': torch.zeros(num_drones, 3).double(),
          'q': torch.tensor([0, 0, 0, 1]).repeat(num_drones, 1).double(),
          'w': torch.zeros(num_drones, 3).double(),
          'wind': torch.zeros(num_drones, 3).double(),
          'rotor_speeds': torch.tensor([init_rotor_speed, init_rotor_speed, init_rotor_speed, init_rotor_speed]).repeat(num_drones, 1).double()}

    # First, we need to make the gym environment. The inputs to the model are as follows...
    env = QuadrotorEnv(num_drones, initial_states=x0, quad_params=quad_params, control_mode=control_abstraction)

    observation = env.reset(options={'initial_state': 'random'})

    # Unpack the observation from the environment
    state = {'x': observation[0:3], 'v': observation[3:6], 'q': observation[6:10], 'w': observation[10:13]}

    # The environment expects the control inputs to all be within the range [-1,1].
    if control_abstraction in ['cmd_vel', 'cmd_acc']:
        n_u = 3
    else:
        n_u = 4
    action = np.random.uniform(-1, 1, (num_drones, n_u))

    # Step forward in the environment
    observation, reward, terminated, truncated = env.step(action)

    # Assert that the output of the environment is as expected. 
    assert type(observation) == np.ndarray
    assert type(reward) == float or type(reward) == np.float64 or type(reward) == np.ndarray

    env.close()

def test_gym_environment():

    print("\nTesting gym environment...")

    # Test the gym environment with the different control abstractions
    print("\tTesting gym environment with control abstractions...")
    control_modes = ['cmd_motor_speeds', 'cmd_motor_thrusts', 'cmd_ctbr', 'cmd_ctbm', 'cmd_vel']
    num_drones = [1, 10]
    for control_mode in control_modes:
        for nd in num_drones:
            print(f"\t\t...control_mode: {control_mode}, num_drones: {nd}")
            run_gym_environment(control_mode, nd)


''' 
Check to confirm that the wind classes adhere to the expected signature.
'''

import inspect
import numpy as np
import glob
import os
import importlib

# The expected signature for the `update` method
EXPECTED_UPDATE_SIGNATURE = {
    'args': ['self', 't', 'position'],  # Expected arguments, 'self' is implicit
    'return_type': np.ndarray  # Expected return type (dict)
}

def check_update_signature(class_obj):
    # Ensure the class has an 'update' method
    assert hasattr(class_obj, 'update'), f"{class_obj.__name__} does not have an 'update' method."

    # Get the signature of the 'update' method
    update_method = getattr(class_obj, 'update')

    # Get the signature of the method using inspect
    signature = inspect.signature(update_method)

    # Get the parameter names and return type
    params = list(signature.parameters.keys())  # Get parameter names as a list

    # Check if the parameter names match
    assert params == EXPECTED_UPDATE_SIGNATURE['args'], f"Expected arguments {EXPECTED_UPDATE_SIGNATURE['args']} for {class_obj.__name__}, but got {params}"

def test_wind_update_method():
    
    # Import all wind from the wind module
    wind_files = glob.glob(os.path.join(os.path.dirname(__file__), '..', 'rotorpy', 'wind', '*.py'))
    wind_objects = []
    print("\nTesting input signatures for all trajectory classes...")
    for file in wind_files:
        wind_name = os.path.basename(file)[:-3]  # Remove '.py' extension
        if wind_name != '__init__': # Ignore the __init__.py file
            print(f"Importing trajectory classes from {wind_name}")
            wind_module = importlib.import_module(f"rotorpy.wind.{wind_name}")  # Import the module
            wind_classes = inspect.getmembers(wind_module, inspect.isclass)            # Get all classes in the module

            # Ignore any dependency classes in the wind module
            wind_objects.extend([wind_class for wind_class in wind_classes if wind_class[1].__module__ == f"rotorpy.wind.{wind_name}"])

    # Iterate over all imported classes from the wind module
    for wind_class in wind_objects:
        print(f"Checking {wind_class[0]} input signatures...")
        if hasattr(wind_class[1], 'update'):  # Check only classes with an 'update' method
            check_update_signature(wind_class[1])



from rotorpy.vehicles.ardupilot_multirotor import Ardupilot
from rotorpy.vehicles.multirotor import Multirotor
from rotorpy.sensors.imu import Imu
from pymavlink import mavutil
import numpy as np
from typing import Tuple

import math

# Constants
R_EARTH = 6378137.0  # meters
rad2deg = 180.0 / np.pi
INT_MAX = 32767
INT_MIN = -32768

class SensorSource:
    """ The binary codes to signal which simulated data is being sent through mavlink

    Credit to Pegasus Simulator https://github.com/PegasusSimulator/PegasusSimulator
    
    Atribute:
        | ACCEL (int): mavlink binary code for the accelerometer (0b0000000000111 = 7)
        | GYRO (int): mavlink binary code for the gyroscope (0b0000000111000 = 56)
        | MAG (int): mavlink binary code for the magnetometer (0b0000111000000=448)
        | BARO (int): mavlink binary code for the barometer (0b1101000000000=6656)
        | DIFF_PRESS (int): mavlink binary code for the pressure sensor (0b0010000000000=1024)
    """

    ACCEL: int = 7    
    GYRO: int = 56          
    MAG: int = 448        
    BARO: int = 6656
    DIFF_PRESS: int = 1024

def _compute_hover_rotor_speeds(mass, k_eta, num_rotors, g=9.81):
    """Solve N·k_eta·ω² = m·g for ω and return an array length num_rotors."""
    omega = np.sqrt((mass * g) / (num_rotors * k_eta))
    return np.full(num_rotors, omega)

class PX4Multirotor(Multirotor):
    """PX4 Multirotor Vehicle Model

    Args:
        quad_params (dict): Quadrotor parameters.
        initial_state (dict, optional): Initial state of the quadrotor.
        autopilot_controller (bool): Whether to use the autopilot controller or not.
    """
    def __init__(
        self,
        quad_params,
        initial_state=None,
        control_abstraction="cmd_motor_speeds",
        aero=True,
        enable_ground=True,
        mavlink_url="tcpin:localhost:4560",
        autopilot_controller=True,
        lockstep=True,
        integrator_kwargs=None
    ):
        integrator_kwargs = integrator_kwargs if integrator_kwargs is not None else {'method':'Radau', 'rtol':1e-3, 'atol':1e-6, 'max_step':0.01}
        # If no initial state passed, initialize to hover at origin
        if initial_state is None:
            initial_state = {
                'x': np.zeros(3),
                'v': np.zeros(3),
                'q': np.array([0, 0, 0, 1]),
                'w': np.zeros(3),
                'wind': np.zeros(3),
                'rotor_speeds': np.zeros(quad_params['num_rotors'])
            }
        super().__init__(
            quad_params=quad_params,
            initial_state=initial_state,
            control_abstraction=control_abstraction,
            aero=aero,
            enable_ground=enable_ground,
            integrator_kwargs=integrator_kwargs
        )
        # Simulated IMU (with noise)
        self.imu = Imu()
        self._enable_imu_noise = True  # Always add a bit of noise to avoid stale detection
        self.t = 0.0

        print("PX4Multirotor: Initializing MAVLink connection... on {}".format(mavlink_url))
        self.conn = mavutil.mavlink_connection(mavlink_url)
        self.conn.wait_heartbeat()
        print("PX4Multirotor: MAVLink connection established.")

        self._autopilot_controller = autopilot_controller
        self._lockstep_enabled = lockstep

    @staticmethod
    def enu_to_geodetic(
        east_m: float,
        north_m: float,
        up_m: float,
        lat0_deg: float = 40.0,
        lon0_deg: float = -74.3,
        alt0_m: float = 0.0,
    ) -> Tuple[float, float, float]:
        """
        Convert local ENU coordinates (meters) to geodetic latitude, longitude, and altitude (WGS-84)
        using the simple equirectangular approximation.

        Assumptions:
        - Small-area approximation (recommended within ~10–20 km of the reference).
        - Spherical Earth with WGS-84 semi-major radius.

        Args:
            lat0_deg (float): Reference latitude (degrees, geodetic).
            lon0_deg (float): Reference longitude (degrees, geodetic).
            alt0_m   (float): Reference altitude above mean sea level (meters).
            east_m   (float): Local ENU 'east' offset from reference (meters).
            north_m  (float): Local ENU 'north' offset from reference (meters).
            up_m     (float): Local ENU 'up' offset from reference (meters).

        Returns:
            (lat_deg, lon_deg, alt_msl_m):
                lat_deg   (float): Latitude in degrees (geodetic).
                lon_deg   (float): Longitude in degrees (geodetic).
                alt_msl_m (float): Altitude above mean sea level in meters.

        Notes:
            - This keeps longitude unwrapped; you can normalize to [-180, 180) if desired.
            - For larger areas or higher accuracy, use a full ENU↔ECEF↔LLA conversion.
        """
        # WGS-84 semi-major axis (meters)
        R = 6378137.0

        # Convert reference latitude to radians once
        lat0_rad = math.radians(lat0_deg)

        # Equirectangular projection back to geodetic
        lat_deg = lat0_deg + (north_m / R) * (180.0 / math.pi)
        lon_deg = lon0_deg + (east_m / (R * math.cos(lat0_rad))) * (180.0 / math.pi)

        # Altitude: ENU 'up' increases MSL altitude
        alt_msl_m = alt0_m + up_m

        # (Optional) normalize longitude to [-180, 180)
        if lon_deg >= 180.0 or lon_deg < -180.0:
            lon_deg = ((lon_deg + 180.0) % 360.0) - 180.0

        return lat_deg, lon_deg, alt_msl_m

    @staticmethod
    def geodetic_to_mavlink(lat_deg: float, lon_deg: float, alt_msl_m: float) -> Tuple[int, int, int]:
        """
        Convert geodetic coordinates to MAVLink integer fields.

        Args:
            lat_deg (float): Latitude in degrees.
            lon_deg (float): Longitude in degrees.
            alt_msl_m (float): Altitude above mean sea level in meters.

        Returns:
            lat_int (int): Latitude in 1E-7 degrees (MAVLink int32).
            lon_int (int): Longitude in 1E-7 degrees (MAVLink int32).
            alt_mm  (int): Altitude above MSL in millimeters (MAVLink int32).
        """
        lat_int = int(round(lat_deg * 1e7))
        lon_int = int(round(lon_deg * 1e7))
        alt_mm = int(round(alt_msl_m * 1000.0))
        return lat_int, lon_int, alt_mm

    def _fetch_latest_px4_control(self, blocking : bool = True):
        """Fetch the latest HIL_ACTUATOR_CONTROLS message from PX4 and update control inputs."""

        msg = self.conn.recv_match(type='HIL_ACTUATOR_CONTROLS', blocking=blocking, timeout=0.01)
        if msg is not None:
            return {'cmd_motor_speeds': [c * self.rotor_speed_max for c in msg.controls[:self.num_rotors]]}

    def _enu_to_ned_cmps(self, v_enu):
        v_n = float(v_enu[1])
        v_e = float(v_enu[0])
        v_d = float(-v_enu[2])
        return (
            int(np.round(v_n * 100.0)),
            int(np.round(v_e * 100.0)),
            int(np.round(v_d * 100.0)),
        )

    def _imu(self, state, statedot):
        meas = self.imu.measurement(state, statedot, with_noise=self._enable_imu_noise)
        a_flu = meas["accel"]
        omega_flu = meas["gyro"]
        # FLU -> FRD
        a_frd = np.array([a_flu[0], -a_flu[1], -a_flu[2]], dtype=float)
        omega_frd = np.array([omega_flu[0], -omega_flu[1], -omega_flu[2]], dtype=float)
        return a_frd, omega_frd

    def _send_hil_state_quaternion(self, state, statedot):
        """
        Send HIL_STATE_QUATERNION message to PX4.
        
        Args:
            state: Current vehicle state
            statedot: State derivative (from the `Multirotor.statedot` method)
        """
        # Convert cartesian ENU position to geodetic coordinates (latitude, longitude and height)
        lat_deg, lon_deg, height_meters = self.enu_to_geodetic(*state['x'])
        lat_e7, lon_e7, alt_mm = self.geodetic_to_mavlink(lat_deg, lon_deg, height_meters)
        
        # Convert quaternion from rotorpy to aerospace convention using ArduPilot's method
        quaternion_flu2ned = Ardupilot._quaternion_rotorpy_to_aerospace(state['q'])
        vx_cms, vy_cms, vz_cms = self._enu_to_ned_cmps(state['v'])

        # Send the ground truth acceleration in the state message (without imu noise)
        a_flu_gt = self.imu.measurement(state, statedot, with_noise=False)["accel"]
        a_frd_gt = np.array([a_flu_gt[0], -a_flu_gt[1], -a_flu_gt[2]], dtype=float)
        a_frd_mg = np.clip(np.round(a_frd_gt / 9.80665 * 1000.0), INT_MIN, INT_MAX).astype(np.int16)

        self.conn.mav.hil_state_quaternion_send(
            int(self.t * 1e6),
            quaternion_flu2ned,
            *tuple(state['w']),
            lat_e7, lon_e7, alt_mm,
            vx_cms, vy_cms, vz_cms,
            0, 0,  # Indicated airspeed, true airspeed
            int(a_frd_mg[0]), int(a_frd_mg[1]), int(a_frd_mg[2])
        )

    def _send_hil_sensor(self, state, statedot):
        """
        Send HIL_SENSOR message to PX4.
        
        Args:
            state: Current vehicle state
            statedot: State derivative (computed externally)
        """
        # Get IMU measurements
        a_ned, omega_ned = self._imu(state, statedot)

        # Only flag accel/gyro as updated (exclude mag and baro-related fields)
        updated_bitmask = SensorSource.ACCEL | SensorSource.GYRO

        self.conn.mav.hil_sensor_send(
            int(self.t * 1e6),
            *tuple(a_ned),
            *tuple(omega_ned),
            *(0.0, 0.0, 0.0),       # Magnetometer (body frame)
            0.0,                    # Abs pressure
            0,                      # Differential pressure
            0.0,                    # Altitude from pressure
            25,                     # Temperature
            fields_updated=updated_bitmask,
        )

    def step(self, state, control, t_step):        
        
        # Compute state derivative once for state and messages
        # and send both HIL messages
        statedot = self.statedot(state, control, 0.0)
        self._send_hil_state_quaternion(state, statedot)
        self._send_hil_sensor(state, statedot)

        # Use PX4 commands only if autopilot_controller is True
        if self._autopilot_controller:
            px4_control = self._fetch_latest_px4_control(blocking=self._lockstep_enabled)
            if px4_control is not None:
                control = px4_control
            else:
                control = {'cmd_motor_speeds': np.zeros(self.num_rotors)}

        else: # In this case we use the control provided by the external controller
            pass
        
        state = super().step(state, control, t_step)
        self.state = state
        self.t += t_step

        return state



''' 
Test utilities that are commonly used in RotorPy. 
''' 

import numpy as np 
import matplotlib.pyplot as plt
import os
from scipy.spatial.transform import Rotation
import copy
import json

# Create a test world data dictionary. 
test_world_data = {
    'bounds': {'extents': [0, 5.0, 0, 2.0, 0, 13.0]},
    'blocks': [
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0], 'color': [1, 0, 0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0], 'color': [1, 0, 0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0], 'color': [1, 0, 0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0], 'color': [1, 0, 0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0]},
        {'extents': [2, 3, 0.0, 2, 0.0, 10.0]}],
    'start': np.array([0, 0, 1]),
    'goal': np.array([4, 0, 2]),
    'resolution': np.array([0.25, 0.25, 0.5]),
    'margin': 0.1,
    'expected_path_length': 20.52
    }

def test_animate():
    from rotorpy.utils.animate import animate
    from rotorpy.world import World
    
    print("\nTesting animate() function")
    
    # Create dummy data. 
    world = World.empty([-1, 1, -1, 1, -1, 1])
    T = np.linspace(0, 1, 100)
    position = np.zeros((T.size, 3))
    rotation = np.eye(3).reshape((1, 3, 3)).repeat(T.size, axis=0)
    wind = np.zeros((T.size, 3))
    
    # Animate the results.
    ani = animate(T, position, rotation, wind=wind, animate_wind=False, world=world, filename=None, blit=False, show_axes=True, close_on_finish=True)

def test_numpy_encoding():
    from rotorpy.utils.numpy_encoding import NumpyJSONEncoder, to_ndarray

    print("\nTesting numpy_encoding.py")

    # Retrieve example data.
    data = copy.deepcopy(test_world_data)

    # Write to a JSON file.
    with open('test_utils.json', 'w') as file:
        file.write(json.dumps(data, cls=NumpyJSONEncoder, indent=4))

    # Read from the JSON file.
    with open('test_utils.json') as file:
        data_out = json.load(file)
    data_out = to_ndarray(data_out)

    # Delete the JSON file.
    os.remove('test_utils.json')

    # Ensure that data_out is the same as data. 
    assert np.allclose(data['bounds']['extents'], data_out['bounds']['extents']), "NumPy array encoding failed."
    assert np.allclose(data['start'], data_out['start']), "NumPy array encoding failed."
    assert np.allclose(data['goal'], data_out['goal']), "NumPy array encoding failed."
    assert np.allclose(data['resolution'], data_out['resolution']), "NumPy array encoding failed."
    assert data['margin'] == data_out['margin'], "NumPy array encoding failed."
    assert data['expected_path_length'] == data_out['expected_path_length'], "NumPy array encoding failed."

def test_occupancy_map():
    from rotorpy.world import World
    from rotorpy.utils.occupancy_map import OccupancyMap

    print("\nTesting occupancy_map.py")

    # Load test world
    world = World(test_world_data)

    # Create a figure
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    # Draw the world
    world.draw(ax)

    # Create an Occupancy map
    oc = OccupancyMap(world, (.5, .5, .5), .01)
    position = np.array([2.5, 2.5, 0.5])
    region = (1, 1)
    oc.get_local_2d_occupancy_map(position, region_size=region)

    # Draw the occupancy map (may be slow for many voxels; will look weird if plotted on top of a world.draw)
    oc.draw(ax)

def test_plotter():
    from rotorpy.utils.animate import animate
    from rotorpy.utils.plotter import Plotter, plot_map
    from rotorpy.world import World

    print("\nTesting plotter.py")

    # Load the test world data.
    world = World(test_world_data)
    
    # Plot the world. 
    (fig, ax) = plt.subplots(nrows=1, ncols=1, num="Top Down World View")
    plot_map(ax, world.world)
    ax.set_title("Test World Plotting")

def test_shapes():
    from rotorpy.utils.shapes import Face, Cuboid, Cylinder, Quadrotor

    print("\nTesting shapes.py")

    # Test Face
    fig = plt.figure(num=4, clear=True)
    ax = fig.add_subplot(projection='3d')
    corners = np.array([(1,1,1), (-1,1,1), (-1,-1,1), (1,-1,1)])
    z_plus_face = Face(ax, corners=corners, facecolors='b')
    x_plus_face = Face(ax, corners=corners, facecolors='r')
    x_plus_face.transform(
        position=(0,0,0),
        rotation=Rotation.from_rotvec(np.pi/2 * np.array([0, 1, 0])).as_matrix())
    y_plus_face = Face(ax, corners=corners, facecolors='g')
    y_plus_face.transform(
        position=(0,0,0),
        rotation=Rotation.from_rotvec(np.pi/2 * np.array([-1, 0, 0])).as_matrix())
    ax.set_xlim(-2,2)
    ax.set_ylim(-2,2)
    ax.set_zlim(-2,2)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    # Test Cuboid
    fig = plt.figure(num=0, clear=True)
    ax = fig.add_subplot(projection='3d')
    cuboid = Cuboid(ax, x_span=1, y_span=1, z_span=1)
    cuboid.transform(position=np.array([[0, 0, 0]]), rotation=np.identity(3))
    rotation = Rotation.from_rotvec(np.pi/4 * np.array([1, 0, 0])).as_matrix()
    cuboid = Cuboid(ax, x_span=1, y_span=2, z_span=3)
    cuboid.transform(position=np.array([[2.0, 1, 1]]), rotation=rotation)
    ax.set_xlim(-1,3)
    ax.set_ylim(-1,3)
    ax.set_zlim(-1,3)

    # Test Cylinder
    fig = plt.figure(num=1, clear=True)
    ax = fig.add_subplot(projection='3d')
    cylinder = Cylinder(ax, radius=0.2, height=0.2)
    cylinder.transform(position=np.array([[0, 0, 0]]), rotation=np.identity(3))
    rotation = Rotation.from_rotvec(np.pi/4 * np.array([1, 0, 0])).as_matrix()
    cylinder = Cylinder(ax, radius=0.2, height=0.2)
    cylinder.transform(position=np.array([[1.0, 0, 0]]), rotation=rotation)
    ax.set_xlim(-1,1)
    ax.set_ylim(-1,1)
    ax.set_zlim(-1,1)

    # Test Quadrotor
    fig = plt.figure(num=2, clear=True)
    ax = fig.add_subplot(projection='3d')
    quad = Quadrotor(ax)
    quad.transform(position=np.array([[0.5, 0.5, 0.5]]), rotation=np.identity(3))
    quad = Quadrotor(ax)
    rotation = Rotation.from_rotvec(np.pi/4 * np.array([1, 0, 0])).as_matrix()
    quad.transform(position=np.array([[0.8, 0.8, 0.8]]), rotation=rotation)
    ax.set_xlim(-1,1)
    ax.set_ylim(-1,1)
    ax.set_zlim(-1,1)

    # Test Cuboid coloring.
    fig = plt.figure(num=3, clear=True)
    ax = fig.add_subplot(projection='3d')
    ax.set_xlim(-3.25,3.25)
    ax.set_ylim(-3.25,3.25)
    ax.set_zlim(-3.25,3.25)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    # No shading.
    x = (-1, 0, 1)
    z = -3.25
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False)
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False, facecolors='b')
    cuboid.transform(position=(x[2], 0, z))

    # Shading.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5)
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, facecolors='b')
    cuboid.transform(position=(x[2], 0, z))

    # Transparency.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5)
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5, facecolors='b')
    cuboid.transform(position=(x[2], 0, z))

    # No shading, edge lines.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False, linewidth=1)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False, linewidth=1, edgecolors='k')
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, shade=False, linewidth=1, facecolors='b', edgecolors='k')
    cuboid.transform(position=(x[2], 0, z))

    # Shading, edge lines.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, linewidth=1)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, linewidth=1, edgecolors='k')
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, linewidth=1, facecolors='b', edgecolors='k')
    cuboid.transform(position=(x[2], 0, z))

    # Transparency, edge lines.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5, linewidth=1)
    cuboid.transform(position=(x[0], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5, linewidth=1, edgecolors='k')
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0.5, linewidth=1, facecolors='b', edgecolors='k')
    cuboid.transform(position=(x[2], 0, z))

    # Transparent edges.
    z = z + 1
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0, linewidth=1, edgecolors='k', antialiased=False)
    cuboid.transform(position=(x[1], 0, z))
    cuboid = Cuboid(ax, x_span=0.5, y_span=0.5, z_span=0.5, alpha=0, linewidth=1, edgecolors='k', antialiased=True)
    cuboid.transform(position=(x[2], 0, z))


"""This class inherits from Multirotor and is used to interface with Ardupilot SITL.
It sends the state to the SITL to verify that the conversion is correct.
"""
import copy
from dataclasses import asdict, dataclass, field
from typing import Dict, List

import numpy as np
from rotorpy.vehicles.multirotor import Multirotor
from rotorpy.sensors.imu import Imu
from rotorpy.vehicles.hummingbird_params import quad_params as crazyflie_params
from scipy.spatial.transform import Rotation as R

try:
    from ArduPilotPlugin import ArduPilotPlugin
except:
    print("Missing dependency ArduPilotPlugin. Install it from source at https://github.com/TomerTip/PyArduPilotPlugin")

PWM_MIN = 1100
PWM_MAX = 1900
SEND_RATE_HZ = 1000

@dataclass
class ControlCommand:
    cmd_motor_speeds: List[int] = field(default_factory=lambda: [0, 0, 0, 0])
        
    def to_dict(self) -> dict:
        return asdict(self)

x0 = {'x': np.array([0,0,0]),
      'v': np.zeros(3,),
      'q': np.array([0, 0, 0, 1]), # [i,j,k,w]
      'w': np.zeros(3,),
      'wind': np.array([0,0,0]),  # Since wind is handled elsewhere, this value is overwritten
      'rotor_speeds': np.array([0.0, 0.0, 0.0, 0.0])}


@dataclass
class SensorData:
    sim_position : List[float]
    sim_attitude : List[float]
    sim_velocity_inertial : List[float]
    xgyro: float
    ygyro: float
    zgyro: float
    xacc: float
    yacc: float 
    zacc: float
    
    def __post_init__(self):
        assert len(self.sim_position) == 3
        assert len(self.sim_velocity_inertial) == 3
        assert len(self.sim_attitude) == 4


class Ardupilot(Multirotor):
    M_glu2frd = R.from_euler('x', np.pi)
    M_enu2ned = R.from_matrix([[0, 1, 0], [1, 0, 0], [0, 0, -1]])

    def __init__(self, quad_params, initial_state = x0, control_abstraction='cmd_motor_speeds', aero=True, ardupilot_control=False, enable_imu_noise = False, enable_ground = False):
        super().__init__(quad_params, initial_state, control_abstraction, aero, enable_ground)
        self._enable_ground = enable_ground
        self._ardupilot_control = ardupilot_control
        statedot = {'vdot' : np.zeros(3,), 'wdot' : np.zeros(3,)}
        self.t = 0.0
        self.imu = Imu()
        self.__enable_imu_noise = enable_imu_noise

        self.sensor_data = self._create_sensor_data(self.initial_state, statedot, self.imu, self.__enable_imu_noise)
        self.ap_link = ArduPilotPlugin()
        self.ap_link.drain_unread_packets()

        self._control_cmd = ControlCommand()   


    def step(self, state, control, t_step):
        received, pwm = self.ap_link.pre_update(self.t)
        if self._ardupilot_control:
            control = {'cmd_motor_speeds': self._motor_cmd_to_omega(self._control_cmd.cmd_motor_speeds)}


        statedot = self.statedot(state, control, t_step)
        state =  super().step(state, control, t_step)
        self.t += t_step

        self.sensor_data = self._create_sensor_data(state, statedot, self.imu, self.__enable_imu_noise)
        
        self.ap_link.post_update(self.sensor_data, self.t)
        if received: # TODO: is this being handled correctly?
            self._control_cmd.cmd_motor_speeds = list(pwm[0:4]) # type: ignore
        
        return state


    @staticmethod
    def _motor_cmd_to_omega(pwm_commands : List[int]) -> List[float]:
        """Convert the pwm commands received from the SITL into angular velocity targets for the motors.

        Args:
            pwm_commands (List[int]): pwm commands [1000-2000] output by Ardupilot SITL

        Returns:
            List[float]: angular velocities targets for the motors.
        """
        rotor_0, rotor_1, rotor_2, rotor_3 = pwm_commands
        reordered_pwm_commands = [rotor_2, rotor_0, rotor_3, rotor_1]
        normalized_commands = [(c-PWM_MIN)/(PWM_MAX-PWM_MIN) for c in reordered_pwm_commands]
        angular_velocities = [838.0*c for c in normalized_commands] # TODO: remove magic constant
        return angular_velocities

    @staticmethod
    def _quaternion_rotorpy_to_aerospace(quaternion_glu2enu: np.ndarray) -> List[float]:
        """
        Convert quaternion from rotorpy convention to aerospace (ArduPilot) convention.
        
        Args:
            quaternion_glu2enu (np.ndarray): Quaternion [x, y, z, w] (scalar-last) representing 
                                           rotation from body frame (GLU) to world frame (ENU)
        
        Returns:
            List[float]: Quaternion [w, x, y, z] (scalar-first) representing rotation from 
                        world frame (NED) to body frame (FRD) - aerospace convention
        
        Notes:
            - Input: rotorpy uses scalar-last [x, y, z, w] for GLU→ENU rotation
            - Output: ArduPilot uses scalar-first [w, x, y, z] for NED→FRD rotation (inverse)
            - Involves coordinate frame transformations: GLU↔FRD and ENU↔NED
        """
        # Convert rotorpy quaternion (GLU→ENU) to rotation object
        R_glu2enu = R.from_quat(quaternion_glu2enu, scalar_first=False)
        
        # Transform to aerospace convention (NED→FRD)
        # R_frd2ned represents the attitude of the FRD frame in the NED frame
        R_frd2ned = Ardupilot.M_enu2ned * R_glu2enu * Ardupilot.M_glu2frd
        
        # Return as scalar-first quaternion [w, x, y, z]
        return R_frd2ned.as_quat(scalar_first=True).tolist()      

    @staticmethod
    def _create_sensor_data(
        state: Dict[str, np.ndarray],
        statedot: Dict[str, np.ndarray],
        imu: Imu,
        enable_imu_noise: bool = False,
    ) -> SensorData:
        """
        Converts the state and state derivative of a `rotorpy.vehicles.multirotor` object to the Ardupilot convention
        for position, velocity, acceleration, quaternions, and angular velocities.
        Args:
            state (Dict[str, np.ndarray]): The current state of the multirotor, including position, velocity, and attitude quaternion.
            statedot (np.ndarray): The state derivative, representing the acceleration.
            imu (Imu): The IMU object used to obtain measurements.
            enable_imu_noise (bool, optional): Flag to enable or disable IMU noise in the measurements. Defaults to False.
        Notes:
            - The attitude quaternion in rotorpy is in scalar-last representation (x, y, z, w) and represents the rotation
              from the body frame (GLU) to the world frame (ENU).
            - The attitude quaternion in Ardupilot is in scalar-first representation (w, x, y, z) and represents the rotation
              from the world frame (NED) to the body frame (FRD).

        Returns:
            SensorData: The sensor data in the Ardupilot convention, including position, velocity, acceleration,
                        attitude quaternion, and angular velocities.
        """

        # 1. Convert quaternion from rotorpy to aerospace convention
        quaternion_aerospace = Ardupilot._quaternion_rotorpy_to_aerospace(state["q"])

        # 2. Obtain the IMU measurements in the GLU frame and transform to FRD
        acceleration = copy.deepcopy(statedot)
        meas_dict = imu.measurement(state, acceleration, with_noise=enable_imu_noise)
        a_glu, omega_glu = meas_dict["accel"], meas_dict["gyro"]
        a_frd = Ardupilot.M_glu2frd.apply(a_glu).tolist()
        omega_frd = Ardupilot.M_glu2frd.apply(omega_glu).tolist()

        return SensorData(
            state["x"].tolist(),
            quaternion_aerospace,
            state["v"].tolist(),
            xgyro=omega_frd[0],
            ygyro=omega_frd[1],
            zgyro=omega_frd[2],
            xacc=a_frd[0],
            yacc=a_frd[1],
            zacc=a_frd[2],
        )


if __name__ == '__main__':
    import time
    r = R.from_euler('y', 0, degrees=True)
    initial_state = {'x': np.array([0,0,0]),
                                            'v': np.zeros(3,),
                                            'q': r.as_quat(scalar_first=False), # [i,j,k,w]
                                            'w': np.zeros(3,),
                                            'wind': np.array([0,0,0]),  # Since wind is handled elsewhere, this value is overwritten
                                            'rotor_speeds': np.array([0, 0, 0, 0])}
    vehicle = Ardupilot(initial_state=initial_state, quad_params = crazyflie_params, ardupilot_control=True, enable_imu_noise=False, enable_ground=True)

    state = initial_state

    dt = 0.01
    while True:
        state = vehicle.step(state, {'cmd_motor_speeds': [0,]*4}, dt)
        time.sleep(dt)



'''
Test the basic sensors to ensure they are functioning as expected.
'''

import numpy as np

IMU_OUTPUT_SIGNATURE = {
    'accel': np.ndarray,  # Acceleration vector (m/s^2)
    'gyro': np.ndarray,  # Angular velocity vector (rad/s)
}

MOCAP_OUTPUT_SIGNATURE = {
    'x': np.ndarray,  # Position vector (m)
    'v': np.ndarray,  # Velocity vector (m/s)
    'q': np.ndarray,  # Quaternion (qx, qy, qz, qw)
    'w': np.ndarray,  # Angular velocity vector (rad/s)
}

def test_imu():
    from rotorpy.sensors.imu import Imu

    print("\nTesting IMU class")

    # Create an instance of the IMU class
    imu_obj = Imu()

    # Check if the class has some important attributes that are used elsewhere
    print("\tChecking for important attributes...")
    params = ['accel_bias', 'gyro_bias', 'accel_variance', 'gyro_variance', 'rate_scale', 'gravity_vector', 'p_BS', 'R_BS']
    for param in params:
        assert hasattr(imu_obj, param), f"IMU class is missing '{param}' attribute"

    # Take one measurement and check if the output is as expected  
    # TODO: This test does not check the accuracy of the IMU model... change p_BS and R_BS and assert that the measurement matches the theoretical value. 
    print("\tChecking if measurement() method works as intended...")
    state = {'x': np.zeros(3), 'v': np.zeros(3), 'q': np.array([0, 0, 0, 1]), 'w': np.zeros(3)}
    acceleration = {'vdot': np.zeros(3), 'wdot': np.zeros(3)}
    measurement = imu_obj.measurement(state, acceleration, with_noise=False)
    assert isinstance(measurement, dict), "measurement() method should return a dictionary"
    for key in IMU_OUTPUT_SIGNATURE.keys():
        assert key in measurement, f"Measurement should have '{key}' key"
        assert isinstance(measurement[key], IMU_OUTPUT_SIGNATURE[key]), f"Measurement[{key}] should be a {IMU_OUTPUT_SIGNATURE[key]}"

def test_external_mocap():
    from rotorpy.sensors.external_mocap import MotionCapture

    print("\nTesting MotionCapture class")

    # Create an instance of the MotionCapture class
    mocap_params={'pos_noise_density':  np.zeros(3),  # noise density for position 
                  'vel_noise_density':  np.zeros(3),          # noise density for velocity
                  'att_noise_density':  np.zeros(3),          # noise density for attitude 
                  'rate_noise_density': np.zeros(3),         # noise density for body rates
                  'vel_artifact_max': 5,                              # maximum magnitude of the artifact in velocity (m/s)
                  'vel_artifact_prob': 0,                         # probability that an artifact will occur for a given velocity measurement
                  'rate_artifact_max': 1,                             # maximum magnitude of the artifact in body rates (rad/s)
                  'rate_artifact_prob': 0                        # probability that an artifact will occur for a given rate measurement
                }
    mocap_obj = MotionCapture(sampling_rate=100, mocap_params=mocap_params, with_artifacts=False)

    # Check if the class has some important attributes that are used elsewhere
    print("\tChecking for important attributes...")
    params = ['rate_scale', 'pos_density', 'vel_density', 'att_density', 'rate_density', 'vel_artifact_max', 'vel_artifact_prob', 'rate_artifact_max', 'rate_artifact_prob', 'initialized', 'with_artifacts']
    for param in params:
        assert hasattr(mocap_obj, param), f"MotionCapture class is missing '{param}' attribute"
    
    # Take one measurement and check if the output is as expected
    print("\tChecking if measurement() method works as intended...")
    state_dict = {'x': np.zeros(3), 'v': np.zeros(3), 'q': np.array([0, 0, 0, 1]), 'w': np.zeros(3)}
    measurement = mocap_obj.measurement(state_dict, with_noise=False, with_artifacts=False)
    assert isinstance(measurement, dict), "measurement() method should return a dictionary"
    for key in MOCAP_OUTPUT_SIGNATURE.keys():
        assert key in measurement, f"Measurement should have '{key}' key"
        assert isinstance(measurement[key], MOCAP_OUTPUT_SIGNATURE[key]), f"Measurement[{key}] should be a {MOCAP_OUTPUT_SIGNATURE[key]}"

    for state in state_dict.keys():
        assert np.allclose(state_dict[state], measurement[state]), f"Measurement for {state} should be the same as the state"

def test_range_sensors():
    from rotorpy.sensors.range_sensors import TwoDRangeSensor
    from rotorpy.world import World

    print("\nTesting TwoDRangeSensor class")

    world = World.empty(extents=[-10, 10, -10, 10, -10, 10])

    # Create an instance of the TwoDRangeSensor class
    range_sensor = TwoDRangeSensor(world, sampling_rate=100, map_resolution=1, angular_resolution=1, angular_fov=360, Dmin=0.025, Dmax=100, noise_density=0.0, fixed_heading=True)

    # Check if the class has some important attributes that are used elsewhere
    print("\tChecking for important attributes...")
    params = ['world', 'occ_map', 'rate_scale', 'angular_resolution', 'angular_fov', 'Dmin', 'Dmax', 'sensor_variance', 'fixed_heading', 'ray_angles', 'N_rays', 'world_edges']
    for param in params:
        assert hasattr(range_sensor, param), f"TwoDRangeSensor class is missing '{param}' attribute"
    
    # Take one measurement and check if the output is as expected
    # TODO: This test does not check the accuracy of the range sensor model... change the world and assert that the measurement matches the theoretical value.
    print("\tChecking if measurement() method works as intended...")
    state_dict = {'x': np.zeros(3), 'v': np.zeros(3), 'q': np.array([0, 0, 0, 1]), 'w': np.zeros(3)}
    measurement = range_sensor.measurement(state_dict, with_noise=False)
    assert isinstance(measurement, np.ndarray), "measurement() method should return a dictionary"


from rotorpy.vehicles.multirotor import BatchedMultirotorParams
from rotorpy.vehicles.crazyflie_params import quad_params as cf_params
import numpy as np

crazyflie_randomizations = {
    "mass": [0.027, 0.033],
    "k_eta": [2.1e-8, 2.5e-8],
}

def generate_random_vehicle_params(num_drones,
                                    device,
                                    nominal_params = cf_params,
                                    randomization_ranges = crazyflie_randomizations):
    """ 
    Generate random vehicle params. 
    Inputs:
        num_drones: the number of drones to generate params for. 
        nominal_params: the nominal parameters for the drone, i.e. the center of each sampling distribution.
        randomization_ranges: the range with which parameters are selected. 
    """

    batch_params = BatchedMultirotorParams([nominal_params for _ in range(num_drones)], num_drones, device)
    for idx in range(num_drones):
        update_vehicle_params(idx,
                              randomization_ranges,
                              batch_params)
    return batch_params

def update_vehicle_params(idx,
                          ranges,
                          params_obj):
    """
    Update vehicle parameters. 
    Inputs:
        idx: the particular idx of the drone. 
        ranges: the range of values to sample from. 
        prams_obj: the object (type BatchedMultirotorParams) to modify.

    """
    min_k_eta = 0
    if "mass" in ranges:
        mass_val = np.random.uniform(ranges["mass"][0], ranges["mass"][1])
        params_obj.update_mass(idx, mass_val)

        # divide rotor speed max by 1.8 to give some margin to allow flight 
        min_k_eta = (mass_val * (params_obj.g / 4) / ((params_obj.rotor_speed_max[idx]/1.2) ** 2)).item()
    if "k_eta" in ranges or "k_m" in ranges or "rotor_pos" in ranges:
        k_eta = np.random.uniform(max(ranges["k_eta"][0], min_k_eta),
                                  ranges["k_eta"][1]) if "k_eta" in ranges else None
        k_m = np.random.uniform(ranges["k_m"][0],
                                ranges["k_m"][1]) if "k_m" in ranges else None
        rotor_pos=None  # not implemented yet
        params_obj.update_thrust_and_rotor_params(idx, k_eta, k_m, rotor_pos)
    if "Ixx" in ranges or "Iyy" in ranges or "Izz" in ranges:
        Ixx = np.random.uniform(ranges["Ixx"][0],
                                ranges["Ixx"][1]) if "Ixx" in ranges else None
        Iyy  = np.random.uniform(ranges["Iyy"][0],
                                 ranges["Iyy"][1]) if "Iyy" in ranges else None
        Izz  = np.random.uniform(ranges["Izz"][0],
                                 ranges["Izz"][1]) if "Izz" in ranges else None
        params_obj.update_inertia(idx, Ixx, Iyy, Izz)
    if "tau_m" in ranges:
        tau_m = np.random.uniform(ranges["tau_m"][0],ranges["tau_m"][1])
        params_obj.tau_m[idx] = tau_m
    if "motor_noise" in ranges:
        motor_noise = np.random.uniform(ranges["motor_noise"][0], ranges["motor_noise"][1])
        params_obj.motor_noise[idx] = motor_noise

    return 


"""
Imports
"""
import numpy as np
import torch

class TrajTemplate(object):
    """
    The trajectory is implemented as a class. There are two required methods for each trajectory class: __init__() and update().
    The __init__() is required for instantiating the class with appropriate parameterizations. For example, if you are doing 
    a circular trajectory you might want to specify the radius of the circle. 
    The update() method is called at each iteration of the simulator. The only input to update is time t. The output of update()
    should be the desired flat outputs in a dictionary, as specified below. 
    """
    def __init__(self):
        """
        This is the constructor for the Trajectory object. A fresh trajectory
        object will be constructed before each mission.
        """

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x    = np.zeros((3,))
        x_dot = np.zeros((3,))
        x_ddot = np.zeros((3,))
        x_dddot = np.zeros((3,))
        x_ddddot = np.zeros((3,))
        yaw    = 0
        yaw_dot = 0
        yaw_ddot = 0

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output

class BatchedTrajTemplate(object):
    """
    This is a batched trajectory template. It's the same as the TrajTemplate class, but it's designed to work with batched environment.
    """
    def __init__(self, M=1):
        """
        This is the constructor for the BatchedTrajectory object. A fresh trajectory
        object will be constructed before each mission.

        Here the parameter M is the number of vehicles in the batch. 
        """

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives for each uav.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x    = torch.zeros((M, 3))
        x_dot = torch.zeros((M, 3))
        x_ddot = torch.zeros((M, 3))
        x_dddot = torch.zeros((M, 3))
        x_ddddot = torch.zeros((M, 3))
        yaw    = torch.zeros((M,))
        yaw_dot = torch.zeros((M,))
        yaw_ddot = torch.zeros((M,))

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output



import numpy as np
import torch
import sys

class ThreeDCircularTraj(object):
    """

    """
    def __init__(self, center=np.array([0,0,0]), radius=np.array([1,1,1]), freq=np.array([0.2,0.2,0.2]), yaw_bool=False):
        """
        This is the constructor for the Trajectory object. A fresh trajectory
        object will be constructed before each mission.

        Inputs:
            center, the center of the circle (m)
            radius, the radius of the circle (m)
            freq, the frequency with which a circle is completed (Hz)
        """

        self.center = center
        self.cx, self.cy, self.cz = center[0], center[1], center[2]
        self.radius = radius
        self.freq = freq

        self.omega = 2*np.pi*self.freq

        self.yaw_bool = yaw_bool

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """
        x        = np.array([self.cx + self.radius[0]*np.cos(self.omega[0]*t),
                             self.cy + self.radius[1]*np.sin(self.omega[1]*t),
                             self.cz + self.radius[2]*np.sin(self.omega[2]*t)])
        x_dot    = np.array([-self.radius[0]*self.omega[0]*np.sin(self.omega[0]*t),
                              self.radius[1]*self.omega[1]*np.cos(self.omega[1]*t),
                              self.radius[2]*self.omega[2]*np.cos(self.omega[2]*t)])
        x_ddot   = np.array([-self.radius[0]*(self.omega[0]**2)*np.cos(self.omega[0]*t),
                             -self.radius[1]*(self.omega[1]**2)*np.sin(self.omega[1]*t),
                             -self.radius[2]*(self.omega[2]**2)*np.sin(self.omega[2]*t)])
        x_dddot  = np.array([ self.radius[0]*(self.omega[0]**3)*np.sin(self.omega[0]*t),
                             -self.radius[1]*(self.omega[1]**3)*np.cos(self.omega[1]*t),
                              self.radius[2]*(self.omega[2]**3)*np.cos(self.omega[2]*t)])
        x_ddddot = np.array([self.radius[0]*(self.omega[0]**4)*np.cos(self.omega[0]*t),
                             self.radius[1]*(self.omega[1]**4)*np.sin(self.omega[1]*t),
                             self.radius[2]*(self.omega[2]**4)*np.sin(self.omega[2]*t)])

        if self.yaw_bool:
            yaw = 0.8*np.pi/2*np.sin(2.5*t)
            yaw_dot = 0.8*2.5*np.pi/2*np.cos(2.5*t)
            yaw_ddot = 0.8*(2.5**2)*np.pi/2*np.sin(2.5*t)
        else:
            yaw = 0.0
            yaw_dot = 0.0
            yaw_ddot = 0.0

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}
        return flat_output

class BatchedThreeDCircularTraj(object):
    """ 
    A batched version of the ThreeDCircularTraj.
    """

    def __init__(self, centers, radii, freqs, yaw_bools, device='cpu'):
        """ 
        This is the batched version of ThreeDCircularTraj. Each input must be a list of the same length, M, 
        where M is the number of circles to be completed.
        Inputs:
            centers, the center of each circle (m)
            radii, the radius of each circle (m)
            freqs, the frequency with which each circle is completed (Hz)
        """

        assert len(centers) == len(radii) == len(freqs) == len(yaw_bools), "BatchedThreeDCircularTraj Error: inputs must be of the same length"
        self.M = len(centers)

        self.centers = torch.tensor(centers, device=device)
        self.radii = torch.tensor(radii, device=device)
        self.freqs = torch.tensor(freqs)
        self.yaw_bools = torch.tensor(yaw_bools, device=device)

        self.omegas = (2*np.pi*self.freqs).to(device)

        return 

    def update(self, t):
        """ 
        Given the present time, return the desired flat output and derivatives.
        """ 

        x = torch.stack([self.centers[:,0] + self.radii[:,0]*torch.cos(self.omegas[:,0]*t),
                            self.centers[:,1] + self.radii[:,1]*torch.sin(self.omegas[:,1]*t),
                            self.centers[:,2] + self.radii[:,2]*torch.sin(self.omegas[:,2]*t)], dim=1)
        x_dot = torch.stack([-self.radii[:,0]*self.omegas[:,0]*torch.sin(self.omegas[:,0]*t),
                            self.radii[:,1]*self.omegas[:,1]*torch.cos(self.omegas[:,1]*t),
                            self.radii[:,2]*self.omegas[:,2]*torch.cos(self.omegas[:,2]*t)], dim=1)
        x_ddot = torch.stack([-self.radii[:,0]*(self.omegas[:,0]**2)*torch.cos(self.omegas[:,0]*t),
                            -self.radii[:,1]*(self.omegas[:,1]**2)*torch.sin(self.omegas[:,1]*t),
                            -self.radii[:,2]*(self.omegas[:,2]**2)*torch.sin(self.omegas[:,2]*t)], dim=1)
        x_dddot = torch.stack([ self.radii[:,0]*(self.omegas[:,0]**3)*torch.sin(self.omegas[:,0]*t),
                            -self.radii[:,1]*(self.omegas[:,1]**3)*torch.cos(self.omegas[:,1]*t),
                             self.radii[:,2]*(self.omegas[:,2]**3)*torch.cos(self.omegas[:,2]*t)], dim=1)
        x_ddddot = torch.stack([self.radii[:,0]*(self.omegas[:,0]**4)*torch.cos(self.omegas[:,0]*t),
                            self.radii[:,1]*(self.omegas[:,1]**4)*torch.sin(self.omegas[:,1]*t),
                            self.radii[:,2]*(self.omegas[:,2]**4)*torch.sin(self.omegas[:,2]*t)], dim=1)
                        
        yaw = 0.8*np.pi/2*np.sin(2.5*t)*self.yaw_bools
        yaw_dot = 0.8*2.5*np.pi/2*np.cos(2.5*t)*self.yaw_bools
        yaw_ddot = 0.8*(2.5**2)*np.pi/2*np.sin(2.5*t)*self.yaw_bools

        flat_output = { 'x':x, 'x_dot':x_dot, 'x_ddot':x_ddot, 'x_dddot':x_dddot, 'x_ddddot':x_ddddot,
                        'yaw':yaw, 'yaw_dot':yaw_dot, 'yaw_ddot':yaw_ddot}

        return flat_output


'''
Tests the examples found in /examples. This is mostly just to check for any broken imports.  
'''
import subprocess
import sys
import os
import pytest
import glob

EXAMPLES_DIR = os.path.join(os.path.dirname(__file__), '..', 'examples')

example_scripts = [
    p for p in glob.glob(os.path.join(EXAMPLES_DIR, "*.py"))
    if os.path.isfile(p) and 'ardupilot' not in os.path.basename(p) and 'ppo' not in os.path.basename(p)
]

@pytest.mark.parametrize("script_path", example_scripts)
def test_example_script_runs(script_path):
    """Test that example scripts run without error or timeout."""
    script_name = os.path.basename(script_path)
    print(f"\nTesting {script_name}")

    env = os.environ.copy()
    env["MPLBACKEND"] = "Agg"

    try:
        result = subprocess.run(
            [sys.executable, script_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=5,
            env=env
        )

        if result.returncode != 0:
            if "EOFError" in result.stderr:
                pytest.skip(f"{script_name} skipped: script waits for user input.")
            else:
                pytest.fail(f"{script_name} failed with error:\n{result.stderr.strip()}")

    except subprocess.TimeoutExpired:
        pytest.skip(f"{script_name} skipped: execution timed out.")



'''
Test the Multirotor RotorPy vehicle class. 
'''

import numpy as np

def test_multirotor_class():
    
    from rotorpy.vehicles.multirotor import Multirotor
    from rotorpy.vehicles.crazyflie_params import quad_params

    print("\nTesting Multirotor class")

    # Create an instance of the Multirotor class
    multirotor_obj = Multirotor(quad_params)

    # Check if the class has some important attributes that are used elsewhere
    print("Checking for important attributes...")
    def check_attribute(obj, attr):
        assert hasattr(obj, attr), f"Multirotor class is missing '{attr}' attribute"
    params = ['mass', 'Ixx', 'Iyy', 'Izz', 'Ixy', 'Ixz', 'Iyz', 
              'c_Dx', 'c_Dy', 'c_Dz', 'num_rotors', 'rotor_pos', 'rotor_dir', 
              'rotor_speed_min', 'rotor_speed_max', 'k_eta', 'k_m', 'k_d', 'k_z', 'k_flap', 
              'tau_m', 'motor_noise', 
              'inertia', 'weight', 'g', 'rotor_drag_matrix', 'drag_matrix']
    for param in params:
        check_attribute(multirotor_obj, param)

    assert hasattr(multirotor_obj, 'rotor_geometry'), "Multirotor class is missing 'rotor_geometry' attribute, is extract_geometry() no longer being called?"

    # Test if step() works. 
    print("Checking if step() method works as intended for all control abstractions...")
    assert hasattr(multirotor_obj, 'step'), "Multirotor class is missing 'step' method"

    current_state = {'x': np.zeros(3), 
                     'v': np.zeros(3), 
                     'q': np.array([0, 0, 0, 1]), 
                     'w': np.zeros(3), 
                     'wind': np.zeros(3), 
                     'rotor_speeds': np.zeros(4)}
    control_input = {'cmd_motor_speeds':np.zeros(4),
                     'cmd_motor_thrusts':np.zeros(4),
                     'cmd_thrust':0.0,
                     'cmd_moment':np.zeros(3),
                     'cmd_q':np.array([0, 0, 0, 1]),
                     'cmd_w':np.zeros(3),
                     'cmd_v':np.zeros(3),
                     'cmd_acc':np.array([1e-5, 1e-5, 1e-5])}  # Cmd acc cannot be exactly 0. 

    for control_abstraction in ['cmd_motor_speeds', 'cmd_motor_thrusts', 'cmd_ctbm', 'cmd_ctbr', 'cmd_ctatt', 'cmd_acc', 'cmd_vel']:
        print(f"\tChecking control abstraction: {control_abstraction}")
        multirotor_obj.control_abstraction = control_abstraction
        new_state = multirotor_obj.step(current_state, control_input, t_step=0.01)
        assert isinstance(new_state, dict), "step() method should return a dictionary"


import numpy as np

d = 0.17 # distance from CoM to rotor (m)

# 10040_sihsim_quadx preset (aligned with PX4 SIH parameters)
quad_params = {
    'mass':             1.0,     # kg      (PX4 param SIH_MASS)
    'Ixx':              0.025,   # kg·m²   (PX4 param SIH_IXX)
    'Iyy':              0.025,   # kg·m²   (PX4 param SIH_IYY)
    'Izz':              0.030,   # kg·m²   (PX4 param SIH_IZZ)
    'Ixy':              0.0,
    'Ixz':              0.0,
    'Iyz':              0.0,

    'num_rotors':       4,
    'rotor_pos': {
        'r1':           d*np.array([ 1.0,  -1.0, 0.0]),
        'r2':           d*np.array([-1.0,   1.0, 0.0]),
        'r3':           d*np.array([ 1.0,   1.0, 0.0]),
        'r4':           d*np.array([-1.0,  -1.0, 0.0]),
    },

    # Sign for each motor’s yaw moment (+ CW, - CCW)
    'rotor_directions': np.array([ -1, -1, 1, 1 ]),
    'rI':               np.array([0.0, 0.0, 0.0]),

    'c_Dx':             1e-2,
    'c_Dy':             1e-2,
    'c_Dz':             1e-2,

    'k_eta':            5e-6,    # This is computed as SIH_T_MAX/(rotor_speed_max^2)
    'k_m':              1e-7,    # max yaw moment per rotor (Nm) to SIH_Q_MAX/(rotor_speed_max^2)
    'k_d':              0.0,    # rotor drag
    'k_z':              0.0,    # induced inflow
    'k_h':              0.0,    # translational lift
    'k_flap':           0.0,    # blade flapping moment

    'tau_m':            0.05,   # Motor response time constant (s) from SIH_T_TAU
    'rotor_speed_min':  0.0,    # [rad/s] zero throttle
    'rotor_speed_max':  1000.0,  # [rad/s] full throttle
    'motor_noise_std':  0.0,
}




import numpy as np
import torch

class BatchedTrajectory(object):
    """
    Constructs a batch of trajectories for multiple uav. Each uav has its own trajectory. 
    This is a brute force way to interface with RotorPy's batched simulation environment. 
    It's likely to be inefficient, but it's a start. 
    If you want to speed things up, you might want to consider implementing a batched/vectorized version of your trajectories.

    """
    def __init__(self, trajectories, device='cpu'):
        """
        Inputs:
            trajectories, a list of trajectory objects, one for each uav. 
        """

        self.trajectories = trajectories
        self.num_uavs = len(trajectories)

        self.x_out = torch.zeros(self.num_uavs, 3, device=device)
        self.x_dot_out = torch.zeros(self.num_uavs, 3, device=device)
        self.x_ddot_out = torch.zeros(self.num_uavs, 3, device=device)
        self.x_dddot_out = torch.zeros(self.num_uavs, 3, device=device)
        self.x_ddddot_out = torch.zeros(self.num_uavs, 3, device=device)
        self.yaw_out = torch.zeros(self.num_uavs, device=device)
        self.yaw_dot_out = torch.zeros(self.num_uavs, device=device)
        self.yaw_ddot_out = torch.zeros(self.num_uavs, device=device)

    def update(self, t):
        """
        Given the present time, return the desired flat output and derivatives for each uav.

        Inputs
            t, time, s
        Outputs
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s
        """

        for i in range(self.num_uavs):
            uav_flat_output = self.trajectories[i].update(t)
            self.x_out[i] = torch.from_numpy(uav_flat_output['x'])
            self.x_dot_out[i] = torch.from_numpy(uav_flat_output['x_dot'])
            self.x_ddot_out[i] = torch.from_numpy(uav_flat_output['x_ddot'])
            self.x_dddot_out[i] = torch.from_numpy(uav_flat_output['x_dddot'])
            self.x_ddddot_out[i] = torch.from_numpy(uav_flat_output['x_ddddot'])
            self.yaw_out[i] = torch.tensor(uav_flat_output['yaw'])
            self.yaw_dot_out[i] = torch.tensor(uav_flat_output['yaw_dot'])
            self.yaw_ddot_out[i] = torch.tensor(uav_flat_output['yaw_ddot'])

        flat_outputs = { 'x':self.x_out, 'x_dot':self.x_dot_out, 'x_ddot':self.x_ddot_out, 'x_dddot':self.x_dddot_out, 'x_ddddot':self.x_ddddot_out,
                        'yaw':self.yaw_out, 'yaw_dot':self.yaw_dot_out, 'yaw_ddot':self.yaw_ddot_out}

        return flat_outputs


from datetime import datetime
from pathlib import Path

import numpy as np
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation

from rotorpy.utils.shapes import Quadrotor

import os

class ClosingFuncAnimation(FuncAnimation):
    def __init__(self, fig, func, *args, **kwargs):
        self._close_on_finish = kwargs.pop('close_on_finish')
        FuncAnimation.__init__(self, fig, func, *args, **kwargs)
        
    def _step(self, *args):
        still_going = FuncAnimation._step(self, *args)
        if self._close_on_finish and not still_going:
            plt.close(self._fig)

def _decimate_index(time, sample_time):
    """
    Given sorted lists of source times and sample times, return indices of
    source time closest to each sample time.
    """
    index = np.arange(time.size)
    sample_index = np.round(np.interp(sample_time, time, index)).astype(int)
    return sample_index

def animate(time, position, rotation, wind, animate_wind, world, filename=None, blit=False, show_axes=True, close_on_finish=False):
    """
    Animate a completed simulation result based on the time, position, and
    rotation history. The animation may be viewed live or saved to a .mp4 video
    (slower, requires additional libraries).

    For a live view, it is absolutely critical to retain a reference to the
    returned object in order to prevent garbage collection before the animation
    has completed displaying.

    Below, M corresponds to the number of drones you're animating. If M is None, i.e. the arrays are (N,3) and (N,3,3), then it is assumed that there is only one drone.
    Otherwise, we iterate over the M drones and animate them on the same axes.

    N is the number of time steps in the simulation.

    Parameters
        time, (N,) with uniform intervals
        position, (N,M,3)
        rotation, (N,M,3,3)
        wind, (N,M,3) world wind velocity
        animate_wind, if True animate wind vector
        world, a World object
        filename, for saved video, or live view if None
        blit, if True use blit for faster animation, default is False
        show_axes, if True plot axes, default is True
        close_on_finish, if True close figure at end of live animation or save, default is False
    """

    # Check if there is only one drone.
    if len(position.shape) == 2:
        position = np.expand_dims(position, axis=1)
        rotation = np.expand_dims(rotation, axis=1)
        wind = np.expand_dims(wind, axis=1)
    M = position.shape[1]

    # Temporal style.
    rtf = 1.0 # real time factor > 1.0 is faster than real time playback
    render_fps = 30

    # Normalize the wind by the max of the wind magnitude on each axis, so that the maximum length of the arrow is decided by the scale factor
    wind_mag = np.max(np.linalg.norm(wind, axis=-1), axis=1)             # Get the wind magnitude time series
    max_wind = np.max(wind_mag)                         # Find the maximum wind magnitude in the time series

    if max_wind != 0:
        wind_arrow_scale_factor = 1                         # Scale factor for the wind arrow
        wind = wind_arrow_scale_factor*wind / max_wind

    # Decimate data to render interval; always include t=0.
    if time[-1] != 0:
        sample_time = np.arange(0, time[-1], 1/render_fps * rtf)
    else:
        sample_time = np.zeros((1,))
    index = _decimate_index(time, sample_time)
    time = time[index]
    position = position[index,:]
    rotation = rotation[index,:]
    wind = wind[index,:]

    # Set up axes.
    if filename is not None:
        if isinstance(filename, Path):
            fig = plt.figure(filename.name)
        else:
            fig = plt.figure(filename)
    else:
        fig = plt.figure('Animation')
    fig.clear()
    ax = fig.add_subplot(projection='3d')
    if not show_axes:
        ax.set_axis_off()

    quads = [Quadrotor(ax, wind=animate_wind, wind_scale_factor=1) for _ in range(M)]

    world_artists = world.draw(ax)

    title_artist = ax.set_title('t = {}'.format(time[0]))

    def init():
        ax.draw(fig.canvas.get_renderer())
        # return world_artists + list(cquad.artists) + [title_artist]
        return world_artists + [title_artist] + [q.artists for q in quads]

    def update(frame):
        title_artist.set_text('t = {:.2f}'.format(time[frame]))
        for i, quad in enumerate(quads):
            quad.transform(position=position[frame,i,:], rotation=rotation[frame,i,:,:], wind=wind[frame,i,:])
        # [a.do_3d_projection(fig.canvas.get_renderer()) for a in quad.artists]   # No longer necessary in newer matplotlib?
        # return world_artists + list(quad.artists) + [title_artist]
        return world_artists + [title_artist] + [q.artists for q in quads]

    ani = ClosingFuncAnimation(fig=fig,
                        func=update,
                        frames=time.size,
                        init_func=init,
                        interval=1000.0/render_fps,
                        repeat=False,
                        blit=blit,
                        close_on_finish=close_on_finish)

    if filename is not None:
        print('Saving Animation')
        if not ".mp4" in filename:
            filename = filename + ".mp4"
        path = os.path.join(os.path.dirname(__file__),'..','data_out',filename)
        ani.save(path,
                 writer='ffmpeg',
                 fps=render_fps,
                 dpi=100)
        if close_on_finish:
            plt.close(fig)
            ani = None

    return ani


import numpy as np
from rotorpy.utils.occupancy_map import OccupancyMap
from rotorpy.trajectories.minsnap import MinSnap

def sample_waypoints(num_waypoints, world, world_buffer=2, check_collision=False, min_distance=1, max_distance=3, max_attempts=1000, start_waypoint=None, end_waypoint=None):
    """
    Samples random waypoints (x,y,z) in the world. Ensures waypoints do not collide with objects, although there is no guarantee that
    the path you generate with these waypoints will be collision free.
    Inputs:
        num_waypoints: Number of waypoints to sample.
        world: Instance of World class containing the map extents and any obstacles.
        world_buffer: Buffer around the world used for sampling. This is used to ensure that waypoints are at least this distance away
            from the edge of the world.
        check_collision: If True, checks for collisions with obstacles. If False, does not check for collisions. Checking collisions slows down the script.
        min_distance: Minimum distance between waypoints consecutive waypoints.
        max_distance: Maximum distance between consecutive waypoints.
        max_attempts: Maximum number of attempts to sample a waypoint.
        start_waypoint: If specified, the first waypoint will always be this point.
        end_waypoint: If specified, the last waypoint will always be this point.
    Outputs:
        waypoints: A list of (x,y,z) waypoints. [[waypoint_1], [waypoint_2], ... , [waypoint_n]]
    """

    if min_distance > max_distance:
        raise Exception("min_distance must be less than or equal to max_distance.")

    def check_distance(waypoint, waypoints, min_distance, max_distance):
        """
        Checks if the waypoint is at least min_distance away from all other waypoints.
        Inputs:
            waypoint: The waypoint to check.
            waypoints: A list of waypoints.
            min_distance: The minimum distance the waypoint must be from all other waypoints.
            max_distance: The maximum distance the waypoint can be from all other waypoints.
        Outputs:
            collision: True if the waypoint is at least min_distance away from all other waypoints. False otherwise.
        """
        collision = False
        for w in waypoints:
            if (np.linalg.norm(waypoint - w) < min_distance) or (
                    np.linalg.norm(waypoint - w) > max_distance
            ):
                collision = True
        return collision

    def check_obstacles(waypoint, occupancy_map):
        """
        Checks if the waypoint is colliding with any obstacles in the world.
        Inputs:
            waypoint: The waypoint to check.
            occupancy_map: An instance of the occupancy map.
        Outputs:
            collision: True if the waypoint is colliding with any obstacles in the world. False otherwise.
        """
        collision = False
        if occupancy_map is not None:
            if occupancy_map.is_occupied_metric(waypoint):
                collision = True
        return collision

    def single_sample(world, current_waypoints, world_buffer, occupancy_map, min_distance, max_distance, max_attempts=1000, rng=None):
        """
        Samples a single waypoint.
        Inputs:
            world: Instance of World class containing the map extents and any obstacles.
            world_buffer: Buffer around the world used for sampling. This is used to ensure that waypoints are at least this distance away
                from the edge of the world.
            occupancy_map: An instance of the occupancy map.
            min_distance: Minimum distance between waypoints consecutive waypoints.
            max_distance: Maximum distance between consecutive waypoints.
            max_attempts: Maximum number of attempts to sample a waypoint.
            rng: Random number generator. If None, uses numpy's random number generator.
        Outputs:
            waypoint: A single (x,y,z) waypoint.
        """

        num_attempts = 0

        world_lower_limits = (
                np.array(world.world["bounds"]["extents"][0::2]) + world_buffer
        )
        world_upper_limits = (
                np.array(world.world["bounds"]["extents"][1::2]) - world_buffer
        )

        if len(current_waypoints) == 0:
            max_distance_lower_limits = world_lower_limits
            max_distance_upper_limits = world_upper_limits
        else:
            max_distance_lower_limits = current_waypoints[-1] - max_distance
            max_distance_upper_limits = current_waypoints[-1] + max_distance

        lower_limits = np.max(
            np.vstack((world_lower_limits, max_distance_lower_limits)), axis=0
        )
        upper_limits = np.min(
            np.vstack((world_upper_limits, max_distance_upper_limits)), axis=0
        )

        waypoint = np.random.uniform(low=lower_limits, high=upper_limits, size=(3,))
        while check_obstacles(waypoint, occupancy_map) or (
                check_distance(waypoint, current_waypoints, min_distance, max_distance)
                if occupancy_map is not None
                else False
        ):
            waypoint = np.random.uniform(low=lower_limits, high=upper_limits, size=(3,))
            num_attempts += 1
            if num_attempts > max_attempts:
                raise Exception(
                    "Could not sample a waypoint after {} attempts. Issue with obstacles: {}, Issue with min/max distance: {}".format(
                        max_attempts,
                        check_obstacles(waypoint, occupancy_map),
                        check_distance(
                            waypoint, current_waypoints, min_distance, max_distance
                        ),
                    )
                )
        return waypoint

    ######################################################################################################################

    waypoints = []

    if check_collision:
        # Create occupancy map from the world. This can potentially be slow, so only do it if the user wants to check for collisions.
        occupancy_map = OccupancyMap(
            world=world, resolution=[0.5, 0.5, 0.5], margin=0.1
        )
    else:
        occupancy_map = None

    if start_waypoint is not None:
        waypoints = [start_waypoint]
    else:
        # Randomly sample a start waypoint.
        waypoints.append(
            single_sample(
                world,
                waypoints,
                world_buffer,
                occupancy_map,
                min_distance,
                max_distance,
                max_attempts,
            )
        )

    num_waypoints -= 1

    if end_waypoint is not None:
        num_waypoints -= 1

    for _ in range(num_waypoints):
        waypoints.append(
            single_sample(
                world,
                waypoints,
                world_buffer,
                occupancy_map,
                min_distance,
                max_distance,
                max_attempts,
            )
        )

    if end_waypoint is not None:
        waypoints.append(end_waypoint)

    return np.array(waypoints)


def sample_trajectory(traj_obj, t_arr):
    """
    Sample a trajectory on a given array of times.
    Inputs:
        traj_obj: this needs to be a RotorPy trajectory object (insofar as it has an update() method that takes a scalar t and returns a dict containing the flat outputs)
        t_arr: an (N_t,) numpy array of times, it's assumed that a value of 0 indicates the start of the trajectory.
    Outputs:
        position: an (N_t, 3) array of the desired positions from the trajectory.
        velocity: an (N_t, 3) array of the desired velocities from the trajectory.
        acceleration: an (N_t, 3) array of the desired accelerations from the trajectory.
        jerk: an (N_t, 3) array of the desired jerks from the trajectory.
        snap: an (N_t, 3) array of the desired snaps from the trajectory.
        yaw: an (N_t,) array of the desired yaw angles from the trajectory.
        yaw_dot: an (N_t,) array of the desired yaw rates from the trajectory.
        yaw_ddot: an (N_t,) array of the desired yaw accelerations from the trajectory.
    """

    N_t = t_arr.shape[0]
    position = np.zeros((N_t, 3))
    velocity = np.zeros((N_t, 3))
    acceleration = np.zeros((N_t, 3))
    jerk = np.zeros((N_t, 3))
    snap = np.zeros((N_t, 3))
    yaw = np.zeros((N_t,))
    yaw_dot = np.zeros((N_t,))
    yaw_ddot = np.zeros((N_t,))

    for i, t in enumerate(t_arr):
        flat = traj_obj.update(t)
        position[i, :] = flat['x']
        velocity[i, :] = flat['x_dot']
        acceleration[i, :] = flat['x_ddot']
        jerk[i, :] = flat['x_dddot']
        snap[i, :] = flat['x_ddddot']
        yaw[i] = flat['yaw']
        yaw_dot[i] = flat['yaw_dot']
        yaw_ddot[i] = flat['yaw_ddot']

    return position, velocity, acceleration, jerk, snap, yaw, yaw_dot, yaw_ddot


def generate_random_minsnap_traj(world,
                                 num_waypoints,
                                 v_avg_des,
                                 min_distance,
                                 max_distance,
                                 start_position,
                                 max_trials=100,
                                 seed=None):
    if seed is not None:
        np.random.seed(seed)
    trial = 0
    while trial < max_trials:
        waypoints = np.array(sample_waypoints(num_waypoints, world, start_waypoint=start_position,
                                              min_distance=min_distance, max_distance=max_distance))
        try:
            traj = MinSnap(waypoints, v_avg=v_avg_des, verbose=False)
            if traj.x_poly is not None and traj.yaw_poly is not None and traj.x_poly.shape==(num_waypoints-1,3,8):
                return traj
        except TypeError:
            trial += 1
    print("Could not generate trajectory! Try decreasing velocity or increasing spacing between waypoints")
    return None
